import logging

from flask import request

from . import scores
from ... import preprocessors, responses, util

logger = logging.getLogger(__name__)


def configure(config, bp, scoring_system):

    precache_map = util.build_precache_map(config)

    @bp.route("/v3/precache/", methods=["POST"])
    @preprocessors.nocache
    @preprocessors.minifiable
    def precache_v3():
        event = request.get_json()
        if event is None:
            return responses.bad_request(
                "Must provide a POST'ed json as an event")

        try:
            score_request = util.build_score_request_from_event(
                precache_map, event)
        except KeyError as e:
            return responses.bad_request(
                "Must provide the '{key}' parameter".format(key=e.args[0]))
        if not score_request:
            return responses.no_content()
        else:
            return scores.process_score_request(score_request)

    return bp



import json
import logging
from urllib.parse import unquote

from flask import request

from . import scores
from ... import preprocessors, responses, util

logger = logging.getLogger(__name__)


def configure(config, bp, scoring_system):

    precache_map = util.build_precache_map(config)

    @bp.route("/v2/precache/", methods=["GET"])
    @preprocessors.nocache
    @preprocessors.minifiable
    def precache_v2():
        if 'event' not in request.args:
            return responses.bad_request(
                "Must provide an 'event' parameter")
        try:
            event = json.loads(unquote(request.args['event']).strip())
        except json.JSONDecodeError:
            return responses.bad_request(
                "Can not parse event argument as JSON blob")

        score_request = util.build_score_request_from_event(precache_map, event)

        if not score_request:
            return responses.no_content()
        else:
            return scores.process_score_request(score_request)

    return bp



import logging

from .scoring_system import ScoringSystem

logger = logging.getLogger(__name__)


class SingleThread(ScoringSystem):

    def _process_missing_scores(self, request, missing_model_set_revs,
                                root_caches, inprogress_results=None):
        rev_scores = {}
        errors = {}

        for missing_models, rev_ids in missing_model_set_revs.items():
            for rev_id in rev_ids:
                if rev_id not in root_caches:
                    continue
                root_cache = root_caches[rev_id]
                try:
                    score_map = self._process_score_map(
                        request, rev_id, missing_models, root_cache)
                    rev_scores[rev_id] = score_map
                except Exception as error:
                    errors[rev_id] = error

        return rev_scores, errors

    @classmethod
    def from_config(cls, config, name, section_key="scoring_systems"):
        logger.info("Loading SingleThread '{0}' from config.".format(name))
        kwargs = cls._kwargs_from_config(
            config, name, section_key=section_key)
        return cls(**kwargs)



from .scoring_system import ScoringSystem
from .single_thread import SingleThread
from .process_pool import ProcessPool
from .celery_queue import CeleryQueue

__all__ = [ScoringSystem, SingleThread, ProcessPool, CeleryQueue]



import os

from flask import render_template, send_file, request

from . import versions
from . import ui
from . import v1
from . import v2
from . import v3

PWD = os.path.dirname(os.path.abspath(__file__))
GEAR_FAVICON = os.path.join(PWD, "../static/favicon/gear/favicon.ico")


def configure(config, bp, score_processor):

    @bp.route("/", methods=["GET"])
    def index():
        return render_template(
            "home.html", **config['ores'].get('home', {}))

    @bp.route("/favicon.ico", methods=["GET"])
    def favicon():
        # Tries to read from config or just loads the gear..
        return send_file(config['ores'].get('favicon', GEAR_FAVICON))

    @bp.app_errorhandler(404)
    def page_not_found(e):
        return render_template(
            '404.html',
            title=request.path,
            host=config['ores']['wsgi']['error_host'],
            alt=config['ores']['wsgi']['error_alt']), 404

    bp = ui.configure(config, bp)
    bp = versions.configure(config, bp)
    bp = v1.configure(config, bp, score_processor)
    bp = v2.configure(config, bp, score_processor)
    bp = v3.configure(config, bp, score_processor)

    return bp



from collections import defaultdict

from ... import util


def format_v1_score_response(response, limit_to_model=None):
    """
    The response format looks like this::

        {
            "<rev_id>": {
                "<model_name>": <score>
                "<model_name>": <score>
            },
            "<rev_id>": {
                "<model_name>": <score>
                "<model_name>": <score>
            }
        }
    """
    response_doc = defaultdict(dict)
    for rev_id, rev_scores in response.scores.items():
        for model_name, score in rev_scores.items():
            response_doc[rev_id][model_name] = score

    for rev_id, rev_errors in response.errors.items():
        for model_name, error in rev_errors.items():
            response_doc[rev_id][model_name] = util.format_error(error)

    if limit_to_model is not None:
        return util.jsonify({rev_id: model_scores[limit_to_model]
                            for rev_id, model_scores in response_doc.items()})
    else:
        return util.jsonify(response_doc)


def format_some_model_info(scoring_system, request, limit_to_model=None):
    scoring_system.check_context_models(request)
    model_infos = {}
    for model_name in request.model_names:
        model_info = \
            scoring_system[request.context_name].format_model_info(
                model_name, fields=[''])
        model_infos[model_name] = model_info

    if limit_to_model is None:
        return util.jsonify({'models': model_infos})
    else:
        return util.jsonify(model_infos[model_name])



import traceback
from collections import defaultdict

from ... import responses, util


def format_v3_score_response(response):
    """
    {
        "<context_name>": {
            "scores": {
                "<rev_id>": {
                    "<model_name>": {
                        "score": <score>,
                        "features": <features>
                    },
                    "<model_name>": {
                        "score": <score>,
                        "features": <features>
                    }
                },
                "<rev_id>": {
                    "<model_name>": {
                        "score": <score>,
                        "features": <features>
                    },
                    "<model_name>": {
                        "score": <score>,
                        "features": <features>
                    }
                }
            },
            "models": {
                "<model_name>": <model_info>,
                "<model_name>": <model_info>
            }
        }
    }
    """
    context_doc = defaultdict(lambda: defaultdict(dict))
    if len(response.scores) > 0 or len(response.errors) > 0:
        for rev_id, rev_scores in response.scores.items():
            for model_name, score in rev_scores.items():
                context_doc['scores'][rev_id][model_name] = \
                    {'score': score}

        for rev_id, rev_errors in response.errors.items():
            for model_name, error in rev_errors.items():
                context_doc['scores'][rev_id][model_name] = \
                    util.format_error(error)

        for rev_id, rev_features in response.features.items():
            for model_name, features in rev_features.items():
                context_doc['scores'][rev_id][model_name]['features'] = \
                    features

    if len(response.model_info) > 0:
        context_doc['models'] = {
            model_name: info_doc
            for model_name, info_doc in response.model_info.items()}

    return util.jsonify({response.context.name: context_doc})


def build_v3_context_model_map(score_request, scoring_system):
    """
    {
        "<context>": {
            "models": {
                "<model_name>": <model_info>,
                "<model_name>": <model_info>
            }
        },
        "<context>": {
            "models": {
                "<model_name>": <model_info>,
                "<model_name>": <model_info>
            }
        }
    }
    """
    try:
        context_models_doc = {}
        for context_name, context in scoring_system.items():
            context_models_doc[context_name] = {'models': {}}
            for model_name in context:
                model_doc = context.format_model_info(
                    model_name, score_request.model_info or ['version'])
                context_models_doc[context_name]['models'][model_name] = model_doc
        return util.jsonify(context_models_doc)
    except Exception:
        return responses.unknown_error(traceback.format_exc())



import importlib
from collections import OrderedDict

from flask import render_template


def configure(config, bp):

    @bp.route("/versions/")
    def versions():
        modules = ['ores', 'revscoring', 'editquality', 'wikiclass',
                   'draftquality']
        versions = OrderedDict()
        for module in modules:
            try:
                versions[module] = importlib.import_module(module).__version__
            except ImportError:
                pass

        return render_template("versions.html", versions=versions)

    return bp



from collections import namedtuple

from nose.tools import eq_

from ...score_request import ScoreRequest
from ..logger import Logger


def test_logger():
    messages = []
    FakeLogger = namedtuple("Logger", ["debug"])
    logging_logger = FakeLogger(lambda m: messages.append(m))

    collector = Logger(logging_logger)
    collector.precache_request(ScoreRequest("foo", [1], {"bar", "derp"}), 100)
    collector.scores_request(ScoreRequest("foo", list(range(50)), {"bar"}), 150)
    collector.datasources_extracted(ScoreRequest("foo", [1], {"bar"}), 10, 25)
    collector.score_processed(ScoreRequest("foo", [1], {"bar"}), 1.1)
    collector.score_timed_out(ScoreRequest("foo", [1], {"bar"}), 15.1)
    collector.score_cache_miss(ScoreRequest("foo", [1], {"derp"}), "derp")
    collector.score_cache_hit(ScoreRequest("foo", [1], {"bar"}), "bar")
    collector.score_errored(ScoreRequest("foo", [1], {"bar"}), "bar")

    eq_(set(messages),
        {"precache_request: foo:{'bar', 'derp'} in 100 seconds",
         "scores_request: foo:{'bar'} for 50 revisions in 150 seconds",
         "datasources_extracted: foo:{'bar'} for 10 revisions in 25 secs",
         "score_processed: foo:{'bar'} in 1.1 seconds",
         "score_timed_out: foo:{'bar'} in 15.1 seconds",
         "score_cache_miss: foo:derp",
         "score_cache_hit: foo:bar",
         "score_errored: foo:bar"},
        set())


def test_from_config():
    # Should throw a socket connection error and no others
    config = {
        'metrics_collectors': {
            'logger': {
                'class': 'ores.metrics_collectors.Logger'
            }
        }
    }
    Logger.from_config(config, 'logger')



import logging
import time
import traceback

import stopit

from .errors import TimeoutError

logger = logging.getLogger(__name__)


def jsonify_error(error):
    error_type = error.__class__.__name__
    message = str(error)

    return {'error': {'type': error_type, 'message': message}}


def timeout(func, *args, seconds=None, **kwargs):
    if seconds is None:
        return func(*args, **kwargs)

    try:
        start = time.time()
        with stopit.SignalTimeout(seconds) as to_ctx_mgr:
            result = func(*args, **kwargs)
        duration = time.time() - start

        # OK, let's check what happened
        if to_ctx_mgr.state == to_ctx_mgr.EXECUTED:
            # All's fine, everything was executed
            return result
        elif to_ctx_mgr.state == to_ctx_mgr.TIMED_OUT:
            # Timeout occurred while executing the block
            raise TimeoutError("Timed out after {0} seconds.".format(duration))
        else:
            traceback.print_tb()
            raise RuntimeError("Something weird happened.")

    except stopit.TimeoutException as e:
        raise TimeoutError("Timed out after {0} seconds.".format(e.seconds))

    duration = time.time() - seconds



from flask.ext.jsonpify import jsonify


def no_content():
    return "", 204


def error(status, code, message):
    return jsonify({'error': {'code': code, 'message': message}}), status


def not_implemented(message=None):
    return error(501, 'not implemented',
                 message or "Route not implemented yet.")


def bad_request(message):
    return error(400, 'bad request', message)


def forbidden(message=None):
    return error(403, 'forbidden',
                 message or "This request requires authentication.")


def not_found(message=None):
    return error(404, 'not found',
                 message or "Nothing found at this location.")


def server_overloaded(message=None):
    return error(503, 'server overloaded',
                 message or ("Cannot process your request because the " +
                             "server is overloaded.  Try again in a" +
                             "few minutes."))


def unknown_error(message):
    return error(500, 'internal server error', message)



import json
import logging

from flask.ext.jsonpify import jsonify as flask_jsonify

from ..score_request import ScoreRequest

logger = logging.getLogger(__name__)


class CacheParsingError(Exception):
    pass


class ParamError(Exception):
    pass


def jsonify(doc):
    return flask_jsonify(normalize_json(doc))


def normalize_json(doc):
    if isinstance(doc, dict):
        return {_ensure_str_key(k): normalize_json(v) for k, v in doc.items()}
    else:
        return doc


def _ensure_str_key(key):
    if isinstance(key, bool):
        return "true" if key else "false"
    else:
        return str(key)


def read_param(request, param, default=None, type=str):
    try:
        value = request.args.get(param, request.form.get(param))
        if value is None:
            return default
        else:
            return type(value)
    except (ValueError, TypeError) as e:
        raise ParamError("Could not interpret {0}. {1}".format(param, str(e)))


def read_bar_split_param(request, param, default=None, type=str):
    values = read_param(request, param, default=default)
    if values is None:
        return []

    try:
        return [type(value) for value in values.split("|")]
    except (ValueError, TypeError) as e:
        raise ParamError("Could not interpret {0}. {1}"
                         .format(param, str(e)))


def format_error(error):
    error_type = error.__class__.__name__
    message = str(error)

    return {'error': {'type': error_type, 'message': message}}


def build_score_request(scoring_system, request, context_name=None, rev_id=None,
                        model_name=None):
    """
    Build an :class:`ores.ScoreRequest` from information contained in a
    request.

    :Parameters:
        scoring_system : :class:`ores.ScoringSystem`
            A scoring system to build request with
        request : :class:`flask.Request`
            A web request to extract information from
        context_name : `str`
            The name of the context to perform scoring
        rev_id : int
            The revision ID to score.  Note that multiple IDs can be provided
            in `request.args`
        model_name = `str`
            The name of the model to score.  Note that multiple models can be
            provided in `request.args`
    """
    rev_ids = parse_rev_ids(request, rev_id)
    model_names = parse_model_names(request, model_name)
    precache = 'precache' in request.args
    include_features = 'features' in request.args
    injection_caches = parse_injection(request, rev_id)
    model_info = parse_model_info(request)

    if context_name and context_name in scoring_system and not model_names:
        model_names = scoring_system[context_name].keys()

    return ScoreRequest(context_name, rev_ids, model_names,
                        precache=precache,
                        include_features=include_features,
                        injection_caches=injection_caches,
                        model_info=model_info)


def parse_rev_ids(request, rev_id):
    if rev_id is not None:
        return [int(rev_id)]
    else:
        return read_bar_split_param(request, "revids", type=int)


def parse_model_names(request, model_name):
    if model_name is not None:
        return [model_name]
    else:
        return read_bar_split_param(request, "models", type=str)


def parse_injection(request, rev_id):
    """Parse values for features / datasources of interest."""
    cache = {}

    if 'inject' in request.values:
        try:
            cache = json.loads(request.values['inject'])
        except json.JSONDecodeError as e:
            raise CacheParsingError(e)

    if rev_id is not None:
        rev_cache = cache

        try:
            for k, v in request.values.items():
                if k.startswith(("feature.", "datasource.")):
                    rev_cache[k] = json.loads(v)
        except json.JSONDecodeError as e:
            raise CacheParsingError(e)

        if len(rev_cache) > 0:
            cache = {rev_id: rev_cache}
    else:
        cache = {int(rev_id): c for rev_id, c in cache.items()}

    return cache or None


def parse_model_info(request):
    return read_bar_split_param(request, "model_info", type=str)


def build_score_request_from_event(precache_map, event):
    context_name = event['database']
    rev_id = event['rev_id']

    # Check to see if we have the context available in our precache_map
    if context_name not in precache_map:
        return None

    # Start building the response document
    event_set = build_event_set(event)

    model_names = {precache_map[context_name][e] for e in event_set
                   if e in precache_map[context_name]}

    if len(model_names) == 0:
        return None

    return ScoreRequest(context_name, [rev_id], model_names)


# TODO: This strategy for building up events is not sustainable.
def build_event_set(event):
    """
    Turn an EventStream event into a set of event types that ORES
    uses internally.
    """
    event_set = set()
    if event['meta']['topic'] == "mediawiki.revision-create":
        event_set.add('edit')

        user_groups = event.get('performer', {}).get('user_groups', [])
        if 'bot' in user_groups:
            event_set.add('bot_edit')
        else:
            event_set.add('nonbot_edit')

        if event['rev_parent_id'] == 0:
            event_set.add('page_creation')
            if 'bot' in user_groups:
                event_set.add('bot_page_creation')
            else:
                event_set.add('nonbot_page_creation')

    return event_set


AVAILABLE_EVENTS = {'edit', 'bot_edit', 'nonbot_edit', 'page_creation',
                    'bot_page_creation', 'nonbot_page_creation'}


def build_precache_map(config):
    """
    Build a mapping of contexts and models from the configuration
    """
    precache_map = {}
    ss_name = config['ores']['scoring_system']
    for context in config['scoring_systems'][ss_name]['scoring_contexts']:
        precache_map[context] = {}
        for model in config['scoring_contexts'][context].get('precache', []):
            precached_config = \
                config['scoring_contexts'][context]['precache'][model]

            events = precached_config['on']
            if len(set(events) - AVAILABLE_EVENTS) > 0:
                logger.warning("{0} events are not available"
                               .format(set(events) - AVAILABLE_EVENTS))
            for event in precached_config['on']:
                precache_map[context][event] = model
                logger.debug("Setting up precaching for {0} in {1} on {2}"
                             .format(model, context, event))

    return precache_map



import logging
from concurrent import futures as cfutures

from ..errors import TimeoutError
from .scoring_system import ScoringSystem

logger = logging.getLogger(__name__)


class ProcessPool(ScoringSystem):

    def __init__(self, *args, workers=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.workers = int(workers) if workers is not None else None

    def _process_missing_scores(self, request, missing_model_set_revs,
                                root_caches, inprogress_results=None):
        rev_scores = {}
        errors = {}

        futures = {}
        with cfutures.ProcessPoolExecutor(max_workers=self.workers) as executor:
            for missing_models, rev_ids in missing_model_set_revs.items():
                for rev_id in rev_ids:
                    if rev_id not in root_caches:
                        continue
                    root_cache = root_caches[rev_id]
                    logger.debug("Submitting _process_score_map for {0}"
                                 .format(request.format(rev_id, missing_models)))
                    future = executor.submit(
                        self._process_score_map, request, rev_id, missing_models,
                        root_cache)
                    futures[rev_id] = future

            for rev_id, future in futures.items():
                try:
                    rev_scores[rev_id] = future.result(timeout=self.timeout)
                except cfutures.TimeoutError:
                    errors[rev_id] = TimeoutError(
                        "Timed out after {0} seconds.".format(self.timeout))
                except Exception as error:
                    errors[rev_id] = error

        return rev_scores, errors

    @classmethod
    def from_config(cls, config, name, section_key="scoring_systems"):
        logger.info("Loading ProcessPool '{0}' from config.".format(name))
        section = config[section_key][name]

        kwargs = cls._kwargs_from_config(
            config, name, section_key=section_key)
        workers = section.get('workers')

        return cls(workers=workers, **kwargs)



import traceback
from collections import defaultdict

from ... import responses, util


def format_v2_score_response(request, response):
    """
    {
        "scores": {
            "<context>": {
                "<model_name>": {
                    "scores": {
                        "<rev_id>": <score>,
                        "<rev_id>": <score>
                    },
                    "features": {
                        "<rev_id>": <features>,
                        "<rev_id>": <features>
                    },
                    "info": <model_info>,
                    "version": "<model_version>"
                }
                "<model_name>": {
                    "scores": {
                        "<rev_id>": <score>,
                        "<rev_id>": <score>
                    },
                    "features": {
                        "<rev_id>": <features>,
                        "<rev_id>": <features>
                    },
                    "info": <model_info>,
                    "version": "<model_version>"
                }
            }
        }
    }
    """
    return util.jsonify({"scores": {
        response.context.name: {
            model_name: format_v2_model(request, response, model_name)
            for model_name in response.request.model_names}}})


def format_v2_model(request, response, model_name):

    model_doc = defaultdict(dict)
    model_doc['version'] = response.context.model_version(model_name)

    if request.model_info and model_name in response.model_info:
        model_doc['info'] = response.model_info[model_name]

    for rev_id, rev_scores in response.scores.items():
        if model_name in rev_scores:
            model_doc['scores'][rev_id] = rev_scores[model_name]

    for rev_id, rev_errors in response.errors.items():
        if model_name in rev_errors:
            model_doc['scores'][rev_id] = \
                util.format_error(rev_errors[model_name])

    for rev_id, rev_features in response.features.items():
        if model_name in rev_features:
                model_doc['features'][rev_id] = rev_features[model_name]

    return model_doc


def build_v2_context_model_map(score_request, scoring_system):
    """
    {
        "scores": {
            "<context>": {
                "<model_name>": {
                    "version": "<model_version>",
                    "info": <model_info>
                }
            },
            "<context>": {
                "<model_name>": {
                    "version": "<model_version>",
                    "info": <model_info>
                }
            }
        }
    """
    try:
        context_models_doc = {}
        for context_name, context in scoring_system.items():
            context_models_doc[context_name] = {}
            for model_name in context:
                model_doc = {'version': context.model_version(model_name)}
                if score_request.model_info:
                    model_doc['info'] = context.format_model_info(
                        model_name, score_request.model_info)
                context_models_doc[context_name][model_name] = model_doc
        return util.jsonify({'scores': context_models_doc})
    except Exception:
        return responses.unknown_error(traceback.format_exc())



import json


class ScoreRequest:
    def __init__(self, context_name, rev_ids, model_names, precache=False,
                 include_features=False, injection_caches=None,
                 model_info=None):
        """
        Construct a ScoreRequest from parameters.

        :Parameters:
            context_name : str
                The name of the content for the query -- usually a wikidb name
            rev_ids : `iterable` ( `int` )
                A set of revision IDs to score
            model_names : `iterable` ( `str` )
                A set of model_names to use in scoring
            precache : bool
                If true, mark the request as a "precache" request
            include_features : bool
                If true, include feature values in the response
            injection_caches : dict
                A mapping of injection_cache to `rev_id` to use for injecting
                cached data when extracting features/scoring.
            model_info : `list` ( `str` )
                A list of model information fields to include in the response
        """
        self.context_name = context_name
        self.rev_ids = set(rev_ids)
        self.model_names = set(model_names)
        self.precache = precache
        self.include_features = include_features
        self.injection_caches = injection_caches or {}
        self.model_info = model_info

    def __str__(self):
        return self.format()

    def format(self, rev_id=None, model_name=None):
        """
        Fomat a request or a sub-part of a request based on a rev_id and/or
        model_name.  This is useful for logging.
        """
        rev_ids = rev_id if rev_id is not None else set(self.rev_ids)
        model_names = model_name if model_name is not None else set(self.model_names)
        common = [self.context_name, rev_ids, model_names]

        optional = []
        if self.precache:
            optional.append("precache")
        if self.include_features:
            optional.append("features")
        if self.injection_caches:
            optional.append("injection_caches={0}".format(self.injection_caches))
        if self.model_info:
            optional.append("model_info=" + json.dumps(self.model_info))

        return "{0}({1})".format(":".join(repr(v) for v in common),
                                 ", ".join(optional))

    def __repr__(self):
        return "{0}({1})".format(
            self.__class__.__name__,
            ", ".join(repr(v) for v in [
                self.context_name,
                self.rev_ids,
                self.model_names,
                "precache={0!r}".format(self.precache),
                "include_features={0!r}".format(self.include_features),
                "injection_caches={0!r}".format(self.injection_caches),
                "model_info={0!r}".format(self.model_info)]))




from importlib import import_module


def import_from_path(path):
    parts = path.split(".")
    module_path = ".".join(parts[:-1])
    attribute_name = parts[-1]

    module = import_module(module_path)

    attribute = getattr(module, attribute_name)

    return attribute



from collections import defaultdict


class ScoreResponse:

    def __init__(self, context, request, model_info=None, scores=None,
                 errors=None, features=None):
        self.context = context
        self.request = request
        self.scores = defaultdict(dict)
        self.errors = defaultdict(dict)
        self.features = defaultdict(dict)
        self.model_info = {}

        scores = scores or []
        errors = errors or []
        features = features or []
        model_info = model_info or []

        for rev_id, model_name, score in scores:
            self.add_score(rev_id, model_name, score)
        for rev_id, model_name, error in scores:
            self.add_error(rev_id, model_name, error)
        for rev_id, model_name, feature_vals in features:
            self.add_features(rev_id, model_name, feature_vals)
        for model_name, info_doc in model_info:
            self.add_model_info(model_name, info_doc)

    def add_score(self, rev_id, model_name, score):
        self.scores[rev_id][model_name] = score

    def add_error(self, rev_id, model_name, error):
        self.errors[rev_id][model_name] = error

    def add_features(self, rev_id, model_name, features):
        self.features[rev_id][model_name] = features

    def add_model_info(self, model_name, info_doc):
        self.model_info[model_name] = info_doc



"""
Runs a pre-caching server against an ORES instance by listening to an
EventStream and submitting requests for scores to the web interface for each
revision as it happens.

If ran by systemd, it uses watchdog to stay alive.

:Usage:
    precached -h | --help
    precached <stream-url> <ores-url> [--config=<path>] [--delay=<secs>]
                                      [--debug]

:Options:
    -h --help        Prints this documentation
    <stream-url>     URL of an RCStream
    <ores-url>       URL of base ORES webserver
    --delay=<secs>   The delay between when an event is received and when the
                     request is sent to ORES. [default: 0]
    --config=<path>  The path to a yaml config file
                     [default: config]
    --debug          Print debugging information
"""
import glob
import json
import logging
import os
import time
from concurrent.futures import ThreadPoolExecutor

import docopt
import requests
import yamlconf
from sseclient import SSEClient

from ..metrics_collectors import MetricsCollector, Null
from .watchdog import notify_socket, watchdog_ping

logger = logging.getLogger(__name__)

MAX_WORKERS = 50


def main(argv=None):
    args = docopt.docopt(__doc__, argv=argv)

    logging.basicConfig(
        level=logging.INFO if not args['--debug'] else logging.DEBUG,
        format='%(asctime)s %(levelname)s:%(name)s -- %(message)s'
    )
    # Requests is loud.  Be quiet requests.
    logging.getLogger("requests").setLevel(logging.WARNING)
    requests.packages.urllib3.disable_warnings()
    # If we're using logging for metrics collection, show it.
    logging.getLogger("ores.metrics_collectors").setLevel(logging.DEBUG)

    stream_url = args['<stream-url>']
    ores_url = args['<ores-url>']
    config_paths = os.path.join(args['--config'], "*.yaml")
    config = yamlconf.load(*(open(p) for p in
                             sorted(glob.glob(config_paths))))
    delay = float(args['--delay'])
    verbose = bool(args['--debug'])
    notify = notify_socket()
    if not notify:
        logger.info('Not being ran as a service, watchdog disabled')
    ss_name = config['ores']['scoring_system']
    if 'metrics_collector' in config['scoring_systems'][ss_name]:
        metrics_collector = MetricsCollector.from_config(
            config, config['scoring_systems'][ss_name]['metrics_collector'])
    else:
        metrics_collector = Null()

    run(stream_url, ores_url, metrics_collector, config, delay,
        notify, verbose)


def run(stream_url, ores_url, metrics_collector, config, delay, notify, verbose):

    # What to do in case of a change
    def precache_a_change(change):
        if delay:
            time.sleep(delay)
        if notify:
            watchdog_ping(*notify)
        start = time.time()
        response = requests.post(ores_url + "/v3/precache/", json=change)
        if response.status == 200:
            logger.info("Scored {0} in {1} seconds."
                        .format(json.dumps(change), round(time.time() - start, 3)))
        elif response.status == 204:
            logger.debug("Nothing to do for {0}".format(json.dumps(change)))
        else:
            logger.error("Scoring {0} and got an error in response:\n{1}"
                         .format(json.dumps(change), response.content))

    # Execute changes!
    with ThreadPoolExecutor(max_workers=MAX_WORKERS) as executor:
        for event in SSEClient(stream_url):
            if event.event == 'message':
                try:
                    change = json.loads(event.data)
                except ValueError:
                    continue

                executor.submit(precache_a_change, change)



import logging
import time
from hashlib import sha1

from revscoring import dependencies
from revscoring import Model
from revscoring.datasources import Datasource
from revscoring.extractors import Extractor
from revscoring.features import trim

logger = logging.getLogger(__name__)


class ScoringContext(dict):
    """
    Represents a context in which scoring can take place.  Usually, a wiki is
    1:1 with a "ScoringContext".

    :Parameters:
        name : str
           The name of the context in which scoring will happen.  This is
           usually a wiki's database name.
        model_map : dict
           A mapping between names and
           :class:`revscoring.ScorerModel`
           instances
        extractor : :class:`revscoring.Extractor`
           An extractor to use for gathering feature values
    """

    def __init__(self, name, model_map, extractor):
        super().__init__()
        self.name = str(name)
        self.update(model_map)
        self.extractor = extractor

    def format_model_info(self, model_name, paths=None):
        model_info = self._get_model_info_for(model_name)
        return model_info.format(paths, formatting="json")

    def format_id_string(self, model_name, rev_id, injection_cache=None):
        version = self.model_version(model_name)
        score_id = ":".join(
            str(v) for v in [self.name, model_name, version, rev_id])
        if injection_cache is None:
            return score_id
        else:
            sorted_tuple = tuple(sorted(injection_cache.items()))
            cache_hash = sha1(bytes(str(sorted_tuple), 'utf8')).hexdigest()
            return score_id + ":" + cache_hash

    def _get_model_info_for(self, model_name):
        return self[model_name].info

    def model_version(self, model_name):
        return self[model_name].version

    def model_features(self, model_name):
        return self[model_name].features

    def process_model_scores(self, model_names, root_cache,
                             include_features=False):
        """
        Generates a score map for a set of models based on a `root_cache`.
        This method performs no substantial IO, but may incur substantial CPU
        usage.

        :Parameters:
            model_names : `set` ( `str` )
                A set of models to score
            root_cache : `dict` ( `str` --> `mixed` )
                A cache of pre-computed root_dependencies for a specific
                revision.  See `extract_root_dependency_caches()`
            include_features : `bool`
                If True, include a map of basic features used in scoring along
                with the model score.  If False, just generate the scores.
        """
        model_scores = {}

        for model_name in model_names:
            model_scores[model_name] = {}

            # Mostly CPU
            model_scores[model_name]['score'] = \
                self._process_score(model_name, dependency_cache=root_cache)

            # Essentially free
            if include_features:
                base_feature_map = self._solve_base_feature_map(
                    model_name, dependency_cache=root_cache)
                model_scores[model_name]['features'] = base_feature_map

        return model_scores

    def _solve_features(self, model_name, dependency_cache=None):
        """
        Solves the vector (`list`) of features for a given model using
        the `dependency_cache` and returns them.
        """
        features = self[model_name].features
        return list(self.extractor.solve(features, cache=dependency_cache))

    def _solve_base_feature_map(self, model_name, dependency_cache=None):
        """
        Solves the leaf :class:`revscoring.Feature` from the dependency for
        `model_name` using `dependency_cache`.  This will return a mapping
        between the `str` name of the base features and the solved values.
        """
        features = list(trim(self[model_name].features))
        feature_values = self.extractor.solve(features, cache=dependency_cache)
        return {str(f): v
                for f, v in zip(features, feature_values)}

    def _process_score(self, model_name, dependency_cache=None):
        """
        Generates a score for a given model using the `dependency_cache`.
        """
        version = self[model_name].version

        start = time.time()
        feature_values = self._solve_features(model_name, dependency_cache)
        logger.debug("Extracted features for {0}:{1}:{2} in {3} secs"
                     .format(self.name, model_name, version,
                             round(time.time() - start, 3)))

        start = time.time()
        score = self[model_name].score(feature_values)
        logger.debug("Scored features for {0}:{1}:{2} in {3} secs"
                     .format(self.name, model_name, version,
                             round(time.time() - start, 3)))

        return score

    def _generate_root_datasources(self, model_names):
        for model_name in model_names:
            for dependency in dependencies.dig(self.model_features(model_name)):
                if isinstance(dependency, Datasource):
                    yield dependency

    def extract_root_dependency_caches(
            self, model_names, rev_ids, injection_caches=None):
        """
        Extracts a mapping of root :class:`revscoring.Datasource`
        capable of generating the features needed for a particular set of
        models without additional IO.  This method implements all of the IO
        heavy operations.  The roots dependency caches produced by calling
        this method can then be passed to `process_model_scores()` for scoring.

        :Parameters:
            model_names : `list` ( `str` )
                The names of a :class:`revscoring.ScorerModel` to
                extract the roots dependencies for
        """
        # Make a copy of dependency_caches
        root_caches = {}
        for rev_id in rev_ids:
            injection_cache = injection_caches.get(rev_id, {}) \
                              if injection_caches is not None else {}
            root_caches[rev_id] = dict(injection_cache.items())

        # Find our root datasources
        root_datasources = \
            list(set(self._generate_root_datasources(model_names)))

        # Extract the root datasource and update the root_caches in-place
        start = time.time()
        error_root_vals = self.extractor.extract(
            rev_ids, root_datasources, caches=root_caches)

        # Check each extraction for errors
        errors = {}
        for rev_id, (error, _) in zip(rev_ids, error_root_vals):
            if error is not None:
                errors[rev_id] = error
                del root_caches[rev_id]
        logger.debug("Extracted root datasources for {0}:{1}:{2} in {3} secs"
                     .format(self.name, set(model_names), rev_ids,
                             round(time.time() - start, 3)))

        # Note that root_caches should have been modified in place
        return root_caches, errors

    @classmethod
    def from_config(cls, config, name, section_key="scoring_contexts"):
        """
        Expects:

            scoring_contexts:
                enwiki:
                    scorer_models:
                        damaging: enwiki_damaging_2014
                        good-faith: enwiki_good-faith_2014
                    extractor: enwiki
                ptwiki:
                    scorer_models:
                        damaging: ptwiki_damaging_2014
                        good-faith: ptwiki_good-faith_2014
                    extractor: ptwiki

            extractors:
                enwiki_api: ...
                ptwiki_api: ...

            scorer_models:
                enwiki_damaging_2014: ...
                enwiki_good-faith_2014: ...
        """
        logger.info("Loading {0} '{1}' from config.".format(cls.__name__, name))
        section = config[section_key][name]

        model_map = {}
        for model_name, key in section['scorer_models'].items():
            scorer_model = Model.from_config(config, key)
            model_map[model_name] = scorer_model

        extractor = Extractor.from_config(config, section['extractor'])

        return cls(name, model_map=model_map, extractor=extractor)


class ClientScoringContext(ScoringContext):
    """
    A simplistic scoring context that is not capable of performing the scoring
    itself.  This ScoringContext is intended to be used in clients where an
    external service actually implements the scoring pattern
    (e.g. :class:`ores.scoring_systems.CeleryQueue`).  This ScoringContext
    saves on unnecessary memory usage, but still provides access to basic
    informational functionality.
    """

    def __init__(self, name, model_map, *args, **kwargs):
        # Load an empty model map
        bare_model_map = {model_name: NotImplemented
                          for model_name, model in model_map.items()}
        super().__init__(name, bare_model_map, *args, **kwargs)

        # Create an info map for use when formatting information
        self.info_map = {model_name: model.info
                         for model_name, model in model_map.items()}
        self.features_map = {model_name: model.features
                             for model_name, model in model_map.items()}

    def _get_model_info_for(self, model_name):
        return self.info_map[model_name]

    def model_version(self, model_name):
        return self.info_map[model_name].get("version")

    def model_features(self, model_name):
        return self.features_map[model_name]

    def process_score(self, *args, **kwargs):
        raise NotImplementedError()



import glob
import os.path
import yamlconf
import logging
import logging.config

DEFAULT_DIRS = ["config/", "/etc/ores/"]


logger = logging.getLogger(__name__)


def build_config(config_dirs=DEFAULT_DIRS,
                 logging_config="logging_config.yaml"):
    # Loads files in alphabetical order based on the bare filename
    config_file_paths = []
    for directory in config_dirs:
        dir_glob = os.path.join(directory, "*.yaml")
        config_file_paths.extend((os.path.basename(path), path)
                                 for path in glob.glob(dir_glob))
    config_file_paths.sort()
    logger.info("Loading configs from {0}".format(config_file_paths))
    config = yamlconf.load(*(open(p) for fn, p in config_file_paths))

    # Load logging config if specified
    if logging_config is not None:
        with open(logging_config) as f:
            logging_config = yamlconf.load(f)
            logging.config.dictConfig(logging_config)

    # Add nltk data path if specified.
    if 'data_paths' in config['ores'] and \
       'nltk' in config['ores']['data_paths']:
        import nltk
        nltk.data.path.append(config['ores']['data_paths']['nltk'])

    return config



"""
Runs a basic test against an ORES api assuming that it supports the standard
configuration "/v2/testwiki/revid/..."

:Usage:
    test_api -h | --help
    test_api <ores-url> [--debug]

:Options:
    -h --help        Prints this documentation
    <ores-url>       URL of base ORES webserver
    --debug          Print debugging information
"""
import json
import logging

import docopt
import requests

logger = logging.getLogger(__name__)


def main(argv=None):
    args = docopt.docopt(__doc__, argv=argv)

    logging.basicConfig(
        level=logging.INFO if not args['--debug'] else logging.DEBUG,
        format='%(asctime)s %(levelname)s:%(name)s -- %(message)s'
    )
    # Requests is loud.  Be quiet requests.
    requests.packages.urllib3.disable_warnings()

    ores_url = args['<ores-url>']
    make_request(ores_url, "/")
    make_request(ores_url, "/ui/")
    make_request(ores_url, "/scores/", is_json=True)
    make_request(ores_url, "/v1/spec/", is_json=True)
    make_request(ores_url, "/v2/spec/", is_json=True)
    make_request(ores_url, "/v3/spec/", is_json=True)
    make_request(
        ores_url, "/v1/scores/testwiki/revid/2342342/", is_json=True,
        equal_to={"2342342": {"prediction": False,
                              "probability": {"false": 0.76, "true": 0.24}}})
    make_request(
        ores_url, "/v2/scores/testwiki/revid/2342342/", is_json=True,
        equal_to={"scores": {"testwiki": {"revid": {"scores": {
            "2342342": {
                "prediction": False,
                "probability": {"false": 0.76, "true": 0.24}
            }
        }, "version": "0.0.0"}}}})
    make_request(
        ores_url, "/v3/scores/testwiki/2342342/revid/", is_json=True,
        equal_to={"testwiki": {
            "models": {"revid": {"version": "0.0.0"}},
            "scores": {"2342342": {
                "revid": {"score": {"prediction": False,
                                    "probability": {"false": 0.76,
                                                    "true": 0.24}}}
            }}}})
    response = requests.get(ores_url + "/404/")
    assert response.status_code == 404, "/404/ didn't get a 404!"

    make_request(
        ores_url,
        "/v3/scores/testwiki/2342342/revid/?features&feature.delay=16",
        is_json=True,
        equal_to={"testwiki": {
            "models": {"revid": {"version": "0.0.0"}},
            "scores": {"2342342": {
                "revid": {"error": {'message': 'Timed out after 15 seconds.',
                                    'type': 'TimeoutError'}}
            }}}})


def make_request(ores_url, path, is_json=False, equal_to=None):
    logger.debug("Requesting {0}".format(path))
    response = requests.get(ores_url + path)
    assert response.status_code == 200, \
           "Non-OK status code {0} for {1}: {2}".format(
               response.status_code, path, response.content)

    content = response.text

    if is_json:
        try:
            content = json.loads(content)
        except ValueError:
            raise RuntimeError("Could not parse the following as JSON: '{0}"
                               .format(content[0:100]))

    if equal_to:
        assert content == equal_to, "{0} != {1}".format(content, equal_to)

    return response



import logging
import socket

from .metrics_collector import MetricsCollector

logger = logging.getLogger(__name__)


class Statsd(MetricsCollector):
    """
    A metrics collector that uses :mod:`statsd` to connect to a graphite
    instance and send incr() and timing() events.
    """

    def __init__(self, statsd_client):
        self.statsd_client = statsd_client

    def precache_request(self, request, duration):
        self.send_timing_event('precache_request', request.context_name,
                               request.model_names, duration=duration)

    def scores_request(self, request, duration):
        self.send_timing_event('scores_request', request.context_name,
                               request.model_names, len(request.rev_ids),
                               duration=duration)
        self.send_increment_event('revision_scored', request.context_name,
                                  request.model_names,
                                  count=len(request.rev_ids))

    def datasources_extracted(self, request, rev_id_count, duration):
        self.send_timing_event('datasources_extracted', request.context_name,
                               request.model_names, rev_id_count,
                               duration=duration)

    def score_processed(self, request, duration):
        self.send_timing_event('score_processed', request.context_name,
                               request.model_names, duration=duration)

    def score_processor_overloaded(self, request):
        self.send_increment_event('score_processor_overloaded',
                                  request.context_name, request.model_names)

    def score_cache_hit(self, request, model_name):
        if request.precache:
            self.send_increment_event(
                'precache_cache_hit', request.context_name, model_name)
        else:
            self.send_increment_event(
                'score_cache_hit', request.context_name, model_name)

    def score_cache_miss(self, request, model_name):
        if request.precache:
            self.send_increment_event(
                'precache_cache_miss', request.context_name, model_name)
        else:
            self.send_increment_event(
                'score_cache_miss', request.context_name, model_name)

    def score_errored(self, request, model_name):
        self.send_increment_event('score_errored', request.context_name,
                                  model_name)

    def score_timed_out(self, request, duration):
        self.send_timing_event('score_timed_out', request.context_name,
                               request.model_names, duration=duration)

    def precache_score(self, request, duration):
        self.send_timing_event('precache_score', request.context_name,
                               request.model_names, duration=duration)

    def precache_scoring_error(self, request, status, duration):
        self.send_timing_event('precache_scoring_error', request.context_name,
                               request.model_names, status, duration=duration)

    def send_timing_event(self, *message_parts, duration=None):
        with self.statsd_client.pipeline() as pipe:
            for message in self.generate_messages(message_parts):
                pipe.timing(message, duration * 1000)

    def send_increment_event(self, *message_parts, count=1):
        with self.statsd_client.pipeline() as pipe:
            for message in self.generate_messages(message_parts):
                pipe.incr(message, count=count)

    @classmethod
    def generate_messages(cls, parts):
        """
        Performs a deterministic first walk of the tree implied by a list of
        message parts.

        generate_messages(["foo", "bar", ["herp", "derp"], "baz"]) yields:
        * "foo"
        * "foo.bar"
        * "foo.bar.herp"
        * "foo.bar.herp.baz"
        * "foo.bar.derp"
        * "foo.bar.derp.baz"
        """
        for message_parts in cls.generate_message_parts(parts):
            yield ".".join(message_parts)

    @classmethod
    def generate_message_parts(cls, parts, i=0, message=[]):
        if len(message) > 0:
            yield message

        if i < len(parts):
            for bpart in cls.branch_message_part(parts[i]):
                yield from cls.generate_message_parts(
                    parts, i=i + 1, message=message + [bpart])

    @classmethod
    def branch_message_part(cls, part):
        if not isinstance(part, str) and \
           hasattr(part, "__iter__"):
            return (str(p) for p in sorted(part))
        else:
            return [str(part)]

    @classmethod
    def from_parameters(cls, *args, **kwargs):
        import statsd
        statsd_client = statsd.StatsClient(*args, **kwargs)
        return cls(statsd_client)

    @classmethod
    def from_config(cls, config, name, section_key="metrics_collectors"):
        """
        metrics_collectors:
            wmflabs_statsd:
                class: ores.metrics_collectors.Statsd
                host: labmon1001.eqiad.wmnet
                prefix: ores.{hostname}
                maxudpsize: 512
        """
        logger.info("Loading Statsd '{0}' from config.".format(name))
        section = config[section_key][name]

        kwargs = {k: v for k, v in section.items() if k != "class"}
        if 'prefix' in kwargs:
            kwargs['prefix'] = kwargs['prefix'] \
                               .format(hostname=socket.gethostname())
        return cls.from_parameters(**kwargs)



"""
This module provides a :class:`ores.api.Session` class that can maintain a
connection to an instance of ORES and efficiently generate scores.

.. autoclass:: ores.api.Session
    :members:
"""
import logging
import time
import urllib.parse
from concurrent.futures import ThreadPoolExecutor

import requests
import requests.adapters
from more_itertools import chunked

logger = logging.getLogger(__name__)


class Session:
    """
    Constructs a session with an ORES API and provides facilities for scoring
    revisions in batch and parallel.

    :Parameters:
        host : str
            The host of ORES to connect to (preceed with http:// or https://)
        user_agent : str
            A User-Agent header to send with every request
        batch_size : int
            The number of scores to batch per request.
        parallel_request : int
            The maximum number of requests to make in parallel
        retries : int
            The maximum number of retries for basic HTTP errors before giving
            up
    """
    DEFAULT_USERAGENT = "ores.api default user-agent"

    def __init__(self, host, user_agent=None, batch_size=50,
                 parallel_requests=4, retries=5):
        self.host = str(host)
        self.batch_size = int(batch_size)
        self.workers = int(parallel_requests)
        self.retries = int(retries)

        self._session = requests.Session()
        self._session.mount(self.host,
                            requests.adapters.HTTPAdapter(max_retries=retries))
        self.headers = {}
        if user_agent is None:
            logger.warning("Sending requests with default User-Agent.  " +
                           "Set 'user_agent' on oresapi.Session to " +
                           "quiet this message.")
            self.headers['User-Agent'] = self.DEFAULT_USERAGENT
        else:
            self.headers['User-Agent'] = user_agent

    def score(self, context, models, revids):
        """
        Genetate scores for model applied to a sequence of revisions.

        :Parameters:
            context : str
                The name of the context -- usually the database name of a wiki
            models : `iterable`
                The names of a models to apply
            revids : `iterable`
                A sequence of revision IDs to score.
        """
        if isinstance(revids, int):
            rev_ids = [revids]
        else:
            rev_ids = [int(rid) for rid in revids]

        return self._score(context, models, rev_ids)

    def _score(self, context, models, rev_ids):
        logging.debug("Starting up thread pool with {0} workers"
                      .format(self.workers))
        with ThreadPoolExecutor(max_workers=self.workers) as executor:
            futures = []
            for rev_id_batch in chunked(rev_ids, self.batch_size):
                rev_id_batch = list(rev_id_batch)
                logging.debug("Starting batch of {0} revids"
                              .format(len(rev_id_batch)))
                futures.append(executor.submit(self._score_request,
                                               context, rev_id_batch,
                                               models))

            for future in futures:
                for score in future.result():
                    yield score

    def _score_request(self, context, rev_ids, models):
        url = self.host + "/v3/scores/{0}/".format(urllib.parse.quote(context))

        params = {'revids': "|".join(str(rid) for rid in rev_ids),
                  'models': "|".join(urllib.parse.quote(model)
                                     for model in models)}
        logging.debug("Sending score request for {0} revisions"
                      .format(len(rev_ids)))
        start = time.time()
        response = requests.get(url, params=params,
                                headers=self.headers,
                                verify=True, stream=True)
        try:
            doc = response.json()
        except ValueError:
            raise RuntimeError("Non-json response: " + response.text[:100])

        logging.debug("Score request completed for " +
                      "{0} revisions completed in {1} seconds"
                      .format(len(rev_ids), round(time.time() - start, 3)))

        if 'error' in doc:
            # TODO: custom class
            raise RuntimeError(doc['error'])

        if 'warnings' in doc:
            for warning_doc in doc['warnings']:
                logger.warn(warning_doc)

        return [doc[context]['scores'][str(rev_id)]
                for rev_id in rev_ids]



import celery

from ..celery_queue import CeleryQueue
from .util import fakewiki


def test_score():
    application = celery.Celery(__name__)
    CeleryQueue(
        {'fakewiki': fakewiki}, application=application, timeout=15)

    # Can't run the following tests because it starts a new thread and that
    # will break our signal timeout strategy.
    # celerytest.start_celery_worker(application, concurrency=1)
    # test_scoring_system(scoring_system)


def test_celery_queue():
    application = celery.Celery(__name__)
    CeleryQueue(
        {'fakewiki': fakewiki}, application=application, timeout=0.10)

    # Can't run the following tests because it starts a new thread and that
    # will break our signal timeout strategy.
    # celerytest.start_celery_worker(application, concurrency=1)
    # response = scoring_system.score(
    #     ScoreRequest("fakewiki", [1], ["fake"],
    #                  injection_caches={1: {wait_time: 0.05}}))
    # assert 1 in response.errors, str(response.errors)
    # assert isinstance(response.errors[1]['fake'], errors.TimeoutError), \
    #        type(response.errors[1]['fake'])



"""
Runs a WSGI-based web server that hosts ORES.

Usage:
    wsgi (-h | --help)
    wsgi [--host=<name>] [--port=<num>] [--config-dir=<path>]...
         [--processes=<num>] [--debug] [--verbose] [--logging-config=<path>]

Options:
    -h --help                Print this documentation
    --host=<name>            The hostname to listen on [default: 0.0.0.0]
    --port=<num>             The port number to start the server on
                             [default: 8080]
    --config-dir=<path>      The path to a directory containing configuration
                             [default: config/]
    --logging-config=<path>  The path to a logging configuration file
    --processes=<num>        The number of parallel processes to handle
                             [default: 16]
    --debug                  Print debug logging information
    --verbose                Print verbose extraction information
"""
import logging

import docopt
from ores.wsgi import server
from ores import ores

from .util import build_config

# This is a hack to help know when the models must or must not be loaded
# into memory.  By setting this to True, ClientScoringContext will be used
# rather than ScoringContext
ores._is_wsgi_client = True


def main(argv=None):
    args = docopt.docopt(__doc__, argv=argv)
    host = args['--host']
    port = int(args['--port'])
    processes = int(args['--processes'])
    verbose = args['--verbose']
    debug = args['--debug']

    run(host, port, processes, verbose, debug,
        config_dirs=args['--config-dir'],
        logging_config=args['--logging-config'])


def run(host, port, processes, verbose, debug, **kwargs):
    logging.basicConfig(
        level=logging.DEBUG if debug else logging.INFO,
        format='%(asctime)s %(levelname)s:%(name)s -- %(message)s'
    )
    logging.getLogger('requests').setLevel(logging.INFO)
    if verbose:
        logging.getLogger('revscoring.dependencies.dependent') \
               .setLevel(logging.DEBUG)
    else:
        logging.getLogger('revscoring.dependencies.dependent') \
               .setLevel(logging.INFO)

    logging.getLogger("ores.metrics_collectors.logger").setLevel(logging.DEBUG)

    application = build(**kwargs)
    application.debug = True
    application.run(host=host, port=port, processes=processes, debug=True)


def build(*args, **kwargs):
    config = build_config(*args, **kwargs)
    return server.configure(config)



from nose.tools import eq_

from ..score_request import ScoreRequest
from ..score_response import ScoreResponse


class FakeContext:

    def __init__(self, name):
        self.name = name


def test_score_request():
    sr = ScoreResponse(
        FakeContext("foowiki"),
        ScoreRequest("foo", [1, 2, 3], ["bar", "baz"]))

    sr.add_score(1, "bar", {"prediction": False})
    sr.add_score(1, "baz", {"prediction": True})
    sr.add_features(1, "baz", {"feature.rev_id": 1})
    sr.add_error(2, "bar", Exception("WTF"))
    sr.add_error(2, "baz", Exception("WTF"))
    sr.add_score(3, "bar", {"prediction": True})
    sr.add_score(3, "baz", {"prediction": False})
    sr.add_features(3, "baz", {"feature.rev_id": 3})

    eq_(sr.scores[3]['baz']['prediction'], False)



import os

import flask_swaggerui
import flask_wikimediaui


def configure(config):

    from flask import Blueprint, Flask

    from . import routes
    from ..scoring_systems import ScoringSystem

    directory = os.path.dirname(os.path.realpath(__file__))

    app = Flask(__name__,
                static_url_path="/BASE_STATIC",
                template_folder=os.path.join(directory, 'templates'))

    app.config['APPLICATION_ROOT'] = config['ores']['wsgi']['application_root']
    app.url_map.strict_slashes = False
    # Configure routes
    bp = Blueprint('ores', __name__,
                   static_folder=os.path.join(directory, 'static'),
                   url_prefix=config['ores']['wsgi']['url_prefix'])

    ss_name = config['ores']['scoring_system']
    scoring_system = ScoringSystem.from_config(config, ss_name)

    bp = routes.configure(config, bp, scoring_system)
    app.register_blueprint(bp)

    # Configure swagger-ui routes
    swagger_bp = flask_swaggerui.build_static_blueprint(
        'ores-swaggerui', __name__,
        url_prefix=config['ores']['wsgi']['url_prefix'])
    app.register_blueprint(swagger_bp)

    # Configure WikimediaUI routes
    wikimedia_bp = flask_wikimediaui.build_static_blueprint(
        'ores-wikimediaui', __name__,
        url_prefix=config['ores']['wsgi']['url_prefix'])
    app.register_blueprint(wikimedia_bp)

    return app



import logging
import re
from itertools import chain
from urllib.parse import urlparse

import celery
import celery.exceptions
import celery.states
import mwapi.errors
import revscoring.errors
from celery.signals import before_task_publish

from .. import errors
from .scoring_system import ScoringSystem

logger = logging.getLogger(__name__)

_applications = []

DEFAULT_CELERY_QUEUE = "celery"
SENT = "SENT"
REQUESTED = "REQUESTED"


@before_task_publish.connect
def update_sent_state(sender=None, body=None, **kwargs):
    for application in _applications:
        task = application.tasks.get(sender)
        backend = task.backend if task else application.backend

        logger.debug("Setting state to {0} for {1}".format(SENT, body['id']))
        backend.store_result(body['id'], result=None, status=SENT)


class CeleryQueue(ScoringSystem):

    def __init__(self, *args, application, queue_maxsize=None, **kwargs):
        super().__init__(*args, **kwargs)
        global _applications
        self.application = application
        self.queue_maxsize = int(queue_maxsize) if queue_maxsize is not None \
                             else None

        self.redis = redis_from_url(self.application.conf.BROKER_URL)

        if self.queue_maxsize is not None and self.redis is None:
            logger.warning("No redis connection.  Can't check queue size")

        self._initialize_tasks()

        _applications.append(application)

    def _initialize_tasks(self):
        expected_errors = (revscoring.errors.RevisionNotFound,
                           revscoring.errors.PageNotFound,
                           revscoring.errors.UserNotFound,
                           revscoring.errors.DependencyError,
                           mwapi.errors.RequestError,
                           mwapi.errors.TimeoutError,
                           errors.TimeoutError)

        @self.application.task(throws=expected_errors,
                               queue=DEFAULT_CELERY_QUEUE)
        def _process_score_map(request, model_names, rev_id, root_cache):
            logger.info("Generating a score map for {0}"
                        .format(request.format(rev_id, model_names)))

            score_map = ScoringSystem._process_score_map(
                self, request, rev_id, model_names,
                root_cache=root_cache)

            logger.info("Completed generating score map for {0}"
                        .format(request.format(rev_id, model_names)))
            return score_map

        @self.application.task(throws=expected_errors,
                               queue=DEFAULT_CELERY_QUEUE)
        def _lookup_score_in_map(result_id, model_name):
            logger.info("Looking up {0} in {1}".format(model_name, result_id))
            score_map_result = self._process_score_map.AsyncResult(result_id)
            try:
                score_map = score_map_result.get(timeout=self.timeout)

            except celery.exceptions.TimeoutError:
                raise errors.TimeoutError(
                    "Timed out after {0} seconds.".format(self.timeout))
            logger.info("Found {0} in {1}!".format(model_name, result_id))
            return score_map[model_name]

        self._process_score_map = _process_score_map
        self._lookup_score_in_map = _lookup_score_in_map

    def _process_missing_scores(self, request, missing_model_set_revs,
                                root_caches, inprogress_results=None):
        logger.debug("Processing missing scores {0}:{1}."
                     .format(request.context_name, missing_model_set_revs))
        context = self[request.context_name]

        inprogress_results = inprogress_results or {}

        # Generate score results
        results = {}
        for missing_models, rev_ids in missing_model_set_revs.items():
            for rev_id in rev_ids:
                injection_cache = request.injection_caches.get(rev_id)
                if rev_id not in root_caches:
                    for model_name in missing_models:
                        task_id = context.format_id_string(
                            model_name, rev_id, injection_cache=injection_cache)
                        self.application.backend.mark_as_failure(
                            task_id, RuntimeError("Never started"))
                    continue
                root_cache = {str(k): v for k, v in root_caches[rev_id].items()}
                result = self._process_score_map.delay(
                    request, missing_models, rev_id, root_cache)

                for model_name in missing_models:
                    task_id = context.format_id_string(
                        model_name, rev_id, injection_cache=injection_cache)
                    score_result = self._lookup_score_in_map.apply_async(
                        args=(result.id, model_name), task_id=task_id,
                        expires=self.timeout)
                    if rev_id in results:
                        results[rev_id][model_name] = score_result
                    else:
                        results[rev_id] = {model_name: score_result}

        # Read results
        rev_scores = {}
        score_errors = {}
        combined_results = chain(inprogress_results.items(), results.items())
        for rev_id, model_results in combined_results:
            if rev_id not in rev_scores:
                rev_scores[rev_id] = {}
            for model_name, score_result in model_results.items():
                try:
                    rev_scores[rev_id][model_name] = \
                        score_result.get(timeout=self.timeout)
                except celery.exceptions.TimeoutError:
                    timeout_error = errors.TimeoutError(
                        "Timed out after {0} seconds.".format(self.timeout))
                    score_errors[rev_id] = timeout_error
                    self.application.backend.mark_as_failure(
                        score_result.id, timeout_error)
                except Exception as error:
                    score_errors[rev_id] = error

        return rev_scores, score_errors

    def _lookup_inprogress_results(self, request, response):
        context = self[request.context_name]

        inprogress_results = {}
        for rev_id in request.rev_ids:
            injection_cache = request.injection_caches.get(rev_id)

            for model_name in request.model_names:
                if rev_id in response.scores and \
                   model_name in response.scores[rev_id]:
                    continue

                task_id = context.format_id_string(
                    model_name, rev_id, injection_cache=injection_cache)
                score_result = \
                    self._lookup_score_in_map.AsyncResult(task_id)
                if score_result.state in (REQUESTED, SENT,
                                          celery.states.STARTED,
                                          celery.states.SUCCESS):
                    logger.info("Found in-progress result for {0} -- {1}"
                                .format(task_id, score_result.state))
                    if rev_id in inprogress_results:
                        inprogress_results[rev_id][model_name] = score_result
                    else:
                        inprogress_results[rev_id] = {model_name: score_result}

        return inprogress_results

    def _register_model_set_revs_to_process(self, request, model_set_revs):
        context = self[request.context_name]

        for model_set, rev_ids in model_set_revs.items():
            for rev_id in rev_ids:
                for model_name in model_set:
                    injection_cache = request.injection_caches.get(rev_id)
                    task_id = context.format_id_string(
                        model_name, rev_id, injection_cache=injection_cache)
                    self.application.backend.store_result(
                        task_id, {}, REQUESTED)

    def _score(self, *args, **kwargs):
        self._check_queue_full()
        return super()._score(*args, **kwargs)

    def _check_queue_full(self):
        # Check redis to see if the queue of waiting tasks is too big.
        # This is a hack to implement backpressure because celery doesn't
        # support it natively.
        # This will result in a race condition, but it should have OK
        # properties.
        if self.redis is not None and self.queue_maxsize is not None:
            queue_size = self.redis.llen(DEFAULT_CELERY_QUEUE)
            if queue_size > self.queue_maxsize:
                message = "Queue size is too full {0}".format(queue_size)
                logger.warning(message)
                raise errors.ScoreProcessorOverloaded(message)

    @classmethod
    def _build_context_map(cls, config, name, section_key="scoring_systems"):
        from .. import ores
        from ..scoring_context import ScoringContext, ClientScoringContext

        section = config[section_key][name]

        if hasattr(ores, "_is_wsgi_client") and ores._is_wsgi_client:
            ScoringContextClass = ClientScoringContext
        else:
            ScoringContextClass = ScoringContext

        return {name: ScoringContextClass.from_config(config, name)
                for name in section['scoring_contexts']}

    @classmethod
    def from_config(cls, config, name, section_key="scoring_systems"):
        logger.info("Loading CeleryQueue '{0}' from config.".format(name))
        section = config[section_key][name]

        kwargs = cls._kwargs_from_config(
            config, name, section_key=section_key)
        queue_maxsize = section.get('queue_maxsize')
        application = celery.Celery(__name__)
        application.conf.update(**{k: v for k, v in section.items()
                                   if k not in ('class', 'context_map',
                                                'score_cache',
                                                'metrics_collector', 'timeout',
                                                'queue_maxsize')})
        application.conf.CELERY_CREATE_MISSING_QUEUES = True

        return cls(application=application,
                   queue_maxsize=queue_maxsize, **kwargs)


PASS_HOST_PORT = re.compile(
    r"(:(?P<password>[^@]+)@)?" +
    r"(?P<host>[^:]+)?" +
    r"(:(?P<port>[0-9]+))?"
)
"""
Matches <password>@<host>:<port>
"""


def redis_from_url(url):
    """
    Converts a redis URL used by celery into a `redis.Redis` object.
    """
    # Makes sure that we only try to import redis when we need
    # to use it
    import redis

    url = url or ""
    parsed_url = urlparse(url)
    if parsed_url.scheme != "redis":
        return None

    kwargs = {}
    match = PASS_HOST_PORT.match(parsed_url.netloc)
    if match.group('password') is not None:
        kwargs['password'] = match.group('password')
    if match.group('host') is not None:
        kwargs['host'] = match.group('host')
    if match.group('port') is not None:
        kwargs['port'] = int(match.group('port'))

    if len(parsed_url.path) > 1:
        # Removes "/" from the beginning
        kwargs['db'] = int(parsed_url.path[1:])

    return redis.StrictRedis(**kwargs)



import time

from nose.tools import eq_, nottest
from revscoring import Extractor, Model
from revscoring.features import Feature
from revscoring.scoring import ModelInfo

from ...score_request import ScoreRequest
from ...scoring_context import ScoringContext

wait_time = Feature("wait_time", returns=float)


def process_wait(wait_time):
    time.sleep(wait_time)
    return wait_time


wait = Feature("wait", process=process_wait, returns=float,
               depends_on=[wait_time])


class FakeSM(Model):

    def __init__(self):
        self.features = [wait]
        self.version = "fake version"
        self.info = ModelInfo()
        self.info['version'] = self.version

    def score(self, feature_values):
        return True


class FakeExtractor(Extractor):

    def extract(self, rev_ids, features, *args, **kwargs):
        return [(None, [0 for feature in features]) for rev_id in rev_ids]


fakewiki = ScoringContext(
    "fakewiki", {'fake': FakeSM(), 'other_fake': FakeSM()}, FakeExtractor())


@nottest
def test_scoring_system(scoring_system):

    response = scoring_system.score(
        ScoreRequest("fakewiki", [1], ["fake"],
                     injection_caches={1: {wait_time: 0.05}},
                     model_info=['version']))
    eq_(response.errors, {})
    eq_(response.scores, {1: {'fake': True}})
    eq_(response.model_info, {'fake': {'version': 'fake version'}})

    response = scoring_system.score(
        ScoreRequest("fakewiki", [1, 2], ["fake", "other_fake"],
                     injection_caches={1: {wait_time: 0.05},
                                       2: {wait_time: 0.01}},
                     model_info=['version']))
    eq_(response.errors, {})
    eq_(response.scores,
        {1: {'fake': True, 'other_fake': True},
         2: {'fake': True, 'other_fake': True}})
    eq_(response.model_info, {'fake': {'version': 'fake version'},
                              'other_fake': {'version': 'fake version'}})



"""
Runs a celery worker process that exposes a task API for
ores.scoring_systems.CeleryQueue

Usage:
    celery [--config-dir=<path>]... [--logging-config=<path>]
           [--debug] [--verbose]

Options:
    -h --help                Prints this documentation
    --config-dir=<path>      The path to a directory containing configuration
                             [default: config/]
    --logging-config=<path>  The path to a logging configuration file
    --debug                  Print debug logging information
    --verbose                Print verbose extraction information
"""
import logging

import docopt

from ..scoring_systems import CeleryQueue
from .util import build_config


def main(argv=None):
    args = docopt.docopt(__doc__, argv=argv)
    verbose = args['--verbose']
    debug = args['--debug']

    run(verbose, debug,
        config_dirs=args['--config-dir'],
        logging_config=args['--logging-config'])


def run(verbose, debug, **kwargs):
    logging.basicConfig(
        level=logging.DEBUG if debug else logging.INFO,
        format='%(asctime)s %(levelname)s:%(name)s -- %(message)s'
    )
    logging.getLogger('requests').setLevel(logging.INFO)
    if verbose:
        logging.getLogger('revscoring.dependencies.dependent') \
               .setLevel(logging.DEBUG)
    else:
        logging.getLogger('revscoring.dependencies.dependent') \
               .setLevel(logging.INFO)

    logging.getLogger("ores.metrics_collectors.logger").setLevel(logging.DEBUG)
    logging.getLogger("stopit").setLevel(logging.ERROR)

    application = build(**kwargs)
    logging.getLogger('ores').setLevel(logging.DEBUG)
    celery_log_level = "DEBUG" if debug else "INFO"
    application.worker_main(
        argv=["celery_worker", "--loglevel=" + celery_log_level])


def build(*args, **kwargs):
    config = build_config(*args, **kwargs)
    scoring_system = CeleryQueue.from_config(
        config, config['ores']['scoring_system'])
    return scoring_system.application



import logging

logger = logging.getLogger(__name__)


class MetricsCollector:
    """
    Collects metrics about usage of ORES.
    """
    def precache_request(self, request, duration):
        raise NotImplementedError()

    def scores_request(self, request, duration):
        raise NotImplementedError()

    def datasources_extracted(self, request, rev_id_count, duration):
        raise NotImplementedError()

    def score_processor_overloaded(self, request):
        raise NotImplementedError()

    def score_processed(self, request, duration):
        raise NotImplementedError()

    def score_cache_hit(self, request, model_name):
        raise NotImplementedError()

    def score_cache_miss(self, request, model_name):
        raise NotImplementedError()

    def score_errored(self, request, model_name):
        raise NotImplementedError()

    def score_timed_out(self, request, duration):
        raise NotImplementedError()

    def precache_scores(self, request, duration):
        raise NotImplementedError()

    def precache_scoring_error(self, request, status, duration):
        raise NotImplementedError()

    @classmethod
    def from_config(cls, config, name, section_key="metrics_collectors"):
        try:
            import yamlconf
        except ImportError:
            raise ImportError("Could not find yamlconf.  This packages is " +
                              "required when using yaml config files.")
        logger.info("Loading MetricsCollector '{0}' from config.".format(name))
        section = config[section_key][name]
        if 'module' in section:
            return yamlconf.import_module(section['module'])
        elif 'class' in section:
            Class = yamlconf.import_module(section['class'])
            return Class.from_config(config, name)



from ... import errors
from ...score_request import ScoreRequest
from ..single_thread import SingleThread
from .util import fakewiki, test_scoring_system, wait_time


def test_score():
    scoring_system = SingleThread({'fakewiki': fakewiki}, timeout=0.10)

    test_scoring_system(scoring_system)


def test_single_thread():
    scoring_system = SingleThread({'fakewiki': fakewiki}, timeout=1)

    response = scoring_system.score(
        ScoreRequest("fakewiki", [1], ["fake"],
                     injection_caches={1: {wait_time: 2}}))
    assert 1 in response.errors, str(response.errors)
    assert isinstance(response.errors[1]['fake'], errors.TimeoutError), \
           type(response.errors[1]['fake'])



from nose.tools import eq_

from ..score_request import ScoreRequest


def test_score_request():
    sr = ScoreRequest("foo", [1, 2, 3], ["bar", "baz"])

    eq_(sr.context_name, "foo")
    eq_(sr.rev_ids, {1, 2, 3})
    eq_(sr.model_names, {"bar", "baz"})
    eq_(sr.precache, False)
    eq_(sr.include_features, False)
    eq_(sr.model_info, None)



import logging

from .metrics_collector import MetricsCollector
from .util import format_set


class Logger(MetricsCollector):
    """
    A metrics collector that uses :mod:`logging` to report usage metrics.
    """
    def __init__(self, logger=None):
        self.logger = logger or logging.getLogger(__name__)

    def precache_request(self, request, duration):
        self.logger.debug(
            "precache_request: {0}:{1} in {2} seconds"
            .format(request.context_name, format_set(request.model_names),
                    round(duration, 3)))

    def scores_request(self, request, duration):
        self.logger.debug(
            "scores_request: {0}:{1} for {2} revisions in {3} seconds"
            .format(request.context_name, format_set(request.model_names),
                    len(request.rev_ids), round(duration, 3)))

    def datasources_extracted(self, request, rev_id_count, duration):
        self.logger.debug(
            "datasources_extracted: {0}:{1} for {2} revisions in {3} secs"
            .format(request.context_name, format_set(request.model_names),
                    rev_id_count, round(duration, 3)))

    def score_processor_overloaded(self, request, count=1):
        self.logger.debug(
            "score_processor_overloaded: " +
            "{0}:{1}".format(request.context_name,
                             format_set(request.model_names)))

    def score_processed(self, request, duration):
        self.logger.debug(
            "score_processed: {0}:{1} in {2} seconds"
            .format(request.context_name, format_set(request.model_names),
                    round(duration, 3)))

    def score_timed_out(self, request, duration):
        self.logger.debug(
            "score_timed_out: {0}:{1} in {2} seconds"
            .format(request.context_name, format_set(request.model_names),
                    round(duration, 3)))

    def score_cache_hit(self, request, model_name):
        if request.precache:
            self.logger.debug("precache_cache_hit: {0}:{1}"
                              .format(request.context_name, model_name))
        else:
            self.logger.debug("score_cache_hit: {0}:{1}"
                              .format(request.context_name, model_name))

    def score_cache_miss(self, request, model_name):
        if request.precache:
            self.logger.debug("precache_cache_miss: {0}:{1}"
                              .format(request.context_name, model_name))
        else:
            self.logger.debug("score_cache_miss: {0}:{1}"
                              .format(request.context_name, model_name))

    def score_errored(self, request, model_name):
        self.logger.debug("score_errored: {0}:{1}"
                          .format(request.context_name, model_name))

    def precache_score(self, request, duration):
        self.logger.debug("precache_score: {0}:{1} in {2} seconds"
                          .format(request.context_name,
                                  format_set(request.model_names),
                                  round(duration, 3)))

    def precache_scoring_error(self, request, status, duration):
        self.logger.debug(
            "precache_scoring_error: {0}:{1} status={2} in {3} seconds"
            .format(request.context_name, format_set(request.model_names),
                    status, round(duration, 3)))

    @classmethod
    def from_config(cls, config, name, section_key="metrics_collectors"):
        """
        metrics_collectors:
            local_logging:
                class: ores.metrics_collectors.Logging
        """
        return cls()



"""
This script provides access to a set of utilities for ORES

* precached -- Starts a daemon that requests scores for revisions as they happen
* score_revisions -- Scores a set of revisions using an ORES API
* stress_test -- Scores a large set of revisions at a configurable rate
* test_api -- Runs a series of tests against a live ORES API

You can also launch a set of production like applications

* applications.wsgi -- A wsgi server
* applications.celery -- A celery worker

{usage}
Options:
    -h | --help  Shows this documentation
    <utility>    The name of the utility to run
"""
import sys
import traceback
from importlib import import_module


USAGE = """Usage:
    {progname} (-h | --help)
    {progname} <utility> [-h | --help]
""".format(progname=sys.argv[0])


def main():

    if len(sys.argv) < 2:
        sys.stderr.write(USAGE)
        sys.exit(1)
    elif sys.argv[1] in ("-h", "--help"):
        sys.stderr.write(__doc__.format(usage=USAGE))
        sys.exit(1)
    elif sys.argv[1][:1] == "-":
        sys.stderr.write(USAGE)
        sys.exit(1)

    module_name = sys.argv[1]

    if module_name.find("application") == 0:
        module_path = "." + module_name
    else:
        module_path = ".utilities." + module_name

    try:
        sys.path.insert(0, ".")
        module = import_module(module_path, package="ores")
    except ImportError:
        sys.stderr.write(traceback.format_exc())
        sys.stderr.write("Could not find module {0}.\n".format(module_path))
        sys.exit(1)

    module.main(sys.argv[2:])



import json
from urllib.parse import urlparse

from flask import jsonify, render_template, request

from ... import preprocessors


def configure(config, bp, score_processor):

    # /spec/
    @bp.route("/v2/spec/", methods=["GET"])
    @preprocessors.nocache
    @preprocessors.minifiable
    def v2_spec():
        return generate_spec(config)

    return bp


def generate_spec(config):
    return jsonify(json.loads(render_template(
        "v2_swagger.json",
        host=urlparse(request.url_root).netloc,
        scheme=config['ores']['wsgi']['scheme'])))



__name__ = "ores"
__version__ = "1.0.0"
__author__ = "Aaron Halfaker"
__author_email__ = "ahalfaker@wikimedia.org"
__description__ = "A webserver for hosting scorer models."
__url__ = "https://github.com/wiki-ai/ores"
__license__ = "MIT"



"""
Runs a stress test against an ORES api by sending it requests at a configurable
rate.

:Usage:
    stress_test -h | --help
    stress_test <context> <ores-url>...
                [--model=<name>]
                [--batch-size=<n>]
                [--delay=<secs>]
                [--input=<path>]
                [--debug]
                [--verbose]

:Options:
    -h --help          Prints this documentation
    <ores-url>         URL of base ORES webserver.  If multiple URLS are
                       provided, they will be cycled through round-robin.
    <context>          The context in which to request scores
    -m --model=<name>  What models to request?  If none are provided, all
                       available models will be scored.
    --batch-size=<n>   The number of revision IDs should included in a single
                       request [default: 1]
    --delay=<secs>     The delay between requests (rpm = 60/delay) [default: 1]
    --input=<path>     The path to a file containing JSON blobs with a 'rev_id'
                       field. [default: <stdin>]
    --debug            Print debugging information
    --verbose          Print dots and stuff to stderr
"""
import logging
import sys
import time
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from itertools import cycle, islice

import docopt
import requests
from revscoring.utilities.util import read_observations

logger = logging.getLogger(__name__)
REQUEST_THREADS = 1000


def main(argv=None):
    args = docopt.docopt(__doc__, argv=argv)

    logging.basicConfig(
        level=logging.INFO if not args['--debug'] else logging.DEBUG,
        format='%(asctime)s %(levelname)s:%(name)s -- %(message)s'
    )
    # Requests is loud.  Be quiet requests.
    requests.packages.urllib3.disable_warnings()

    ores_urls = args['<ores-url>']
    context = args['<context>']
    if args['--input'] == "<stdin>":
        rev_ids = [ob['rev_id'] for ob in read_observations(sys.stdin)]
    else:
        rev_ids = [
            ob['rev_id'] for ob in read_observations(open(args['--input']))]
    if args['--model'] is None:
        models = []
    else:
        models = args['--model']

    batch_size = int(args['--batch-size'])
    delay = float(args['--delay'])
    verbose = args['--verbose']

    run(ores_urls, context, models, rev_ids, batch_size, delay, verbose)


def run(ores_urls, context, models, rev_ids, batch_size, delay, verbose):
    url_cycle = cycle(ores_urls)
    rev_ids_cycle = cycle(rev_ids)
    stats = defaultdict(int)
    stats["requests"] = defaultdict(int)
    stats["responses"] = defaultdict(int)
    stats["model_scored"] = defaultdict(int)
    stats["model_errored"] = defaultdict(lambda: defaultdict(int))

    start = time.time()
    with ThreadPoolExecutor(max_workers=REQUEST_THREADS) as executor:
        try:
            while time.sleep(delay) is None:
                ores_url = next(url_cycle)
                rev_id_batch = list(islice(rev_ids_cycle, batch_size))
                logger.debug("Submitting {0} rev_ids to {1}...".format(
                    len(rev_ids), ores_url))
                executor.submit(request_and_stat, ores_url, context,
                                models, rev_id_batch, stats, verbose)
        except KeyboardInterrupt:
            print("Caught keyboard interrupt...")
            print("Stats:")
            duration = time.time() - start
            print("  Duration: {0}s".format(round(duration, 3)))
            print("  Requests: {0} ({1} per minute)".format(
                stats['total_requests'],
                round(stats['total_requests'] / (duration / 60), 3)))
            print("  Responses:", stats['total_responses'])
            for status_code, count in stats['responses'].items():
                print("   - {0}: {1}".format(status_code, count))
            print("  Score requests:")
            print("   - errored:", stats['score_requests_errored'])
            print("   - returned:", stats['score_requests_returned'])
            print("  Model scores:")
            print("   - successes:")
            for name, count in stats['model_scored'].items():
                print("     - {0}: {1}".format(name, count))
            print("   - errored:")
            for name, type_doc in stats['model_errored'].items():
                print("     - {0}:".format(name))
                for type, count in type_doc.items():
                    print("       - {0}: {1}".format(type, count))


def request_and_stat(ores_url, context, models, rev_ids, stats, verbose):
    path = "/v3/scores/{0}/".format(context)
    logger.debug("Requesting {0} rev_ids...".format(len(rev_ids)))
    params = {'revids': "|".join(str(r) for r in rev_ids),
              'features': ""}
    if models:
        params = {'models': "|".join(models)}

    stats['total_requests'] += 1
    try:
        response = requests.get(
            ores_url + path, params=params)
    except:
        stats['responses']['connection error'] += 1
        return

    logger.debug("{0} response for {1} rev_ids...".format(
        response.status_code, len(rev_ids)))
    stats['total_responses'] += 1
    stats['responses'][response.status_code] += 1

    if response.status_code >= 200 and response.status_code < 300:
        sys.stderr.write(".")
        sys.stderr.flush()
    elif response.status_code == 503:
        sys.stderr.write("!")
        sys.stderr.flush()
    else:
        sys.stderr.write("?")
        sys.stderr.flush()

    try:
        doc = response.json()
    except ValueError:
        stats['non_json_responses'] += 1
        return

    if 'error' in doc:
        stats['score_requests_errored'] += 1
    elif context in doc:
        stats['score_requests_returned'] += 1

        for rev_id_str in doc[context]['scores']:
            stats['revisions_scored'] += 1
            for name, score_doc in doc[context]['scores'][rev_id_str].items():
                if 'error' in score_doc:
                    type_name = score_doc['error']['type']
                    stats['model_errored'][name][type_name] += 1
                else:
                    stats['model_scored'][name] += 1

    return



from collections import namedtuple

from nose.tools import eq_
from revscoring import dependencies

from ..scoring_context import ScoringContext


def test_scoring_context():
    from revscoring.datasources import Datasource
    from revscoring.dependencies import Dependent
    from revscoring.features import Feature

    fake_data = Datasource("fake_data", lambda: "fake")
    len_func = Dependent("len_func")
    literal_fake = Dependent("literal_fake")
    characters = Feature("characters", lambda word, len: len(word),
                         returns=int,
                         depends_on=[fake_data, len_func])
    is_fake = Feature("is_fake", lambda word, fake: word == fake,
                      returns=bool,
                      depends_on=[fake_data, literal_fake])

    FakeExtractor = namedtuple("Extractor", ['extract', 'solve', 'language'])

    def fake_extract(rev_ids, dependents, caches=None):
        caches = caches if caches is not None else {}
        for rev_id in rev_ids:
            if rev_id % 5 != 0:
                cache = caches.get(rev_id, {})
                values = dependencies.solve(dependents,
                                            context={len_func: lambda: len},
                                            cache=cache)
                values = list(values)
                caches[rev_id] = cache
                yield None, values
            else:
                yield RuntimeError("extract"), None

    def fake_solve(dependents, cache=None):
        cache = cache if cache is not None else {}
        cache.update({len_func: len,
                      literal_fake: "fake"})
        return dependencies.solve(dependents, cache=cache)

    extractor = FakeExtractor(fake_extract, fake_solve, None)

    FakeScorerModel = namedtuple("FakeScorerModel",
                                 ['score', 'version', 'language', 'features'])
    scorer_model = FakeScorerModel(lambda fvs: {"prediction": "generated"},
                                   "1", None, [characters, is_fake])

    scoring_context = ScoringContext("fakewiki", {"fake": scorer_model},
                                     extractor)

    rev_ids = [1, 2, 3, 4, 5]
    root_ds_caches, errors = scoring_context.extract_root_dependency_caches(
        ["fake"], rev_ids)
    print(root_ds_caches)
    print(errors)
    eq_(len(root_ds_caches), 4)
    eq_(len(errors), 1)
    eq_(root_ds_caches[1][fake_data], "fake")
    assert 5 in errors

    score = scoring_context.process_model_scores(
        ["fake"], {characters: 10, is_fake: False})
    eq_(score['fake']['score']['prediction'], "generated")



import traceback

from flask import request

from . import util
from ... import preprocessors, responses
from .... import errors
from ...util import build_score_request


def configure(config, bp, scoring_system):

    # /v2/scores/
    @bp.route("/v2/scores/", methods=["GET"])
    @preprocessors.nocache
    @preprocessors.minifiable
    def scores_v2():
        try:
            score_request = build_score_request(scoring_system, request)
        except Exception as e:
            return responses.bad_request(str(e))

        return util.build_v2_context_model_map(score_request, scoring_system)

    def process_score_request(score_request):
        try:
            score_response = scoring_system.score(score_request)
            return util.format_v2_score_response(score_request, score_response)
        except errors.ScoreProcessorOverloaded:
            return responses.server_overloaded()
        except errors.MissingContext as e:
            return responses.not_found("No scorers available for {0}"
                                       .format(e))
        except errors.MissingModels as e:
            context_name, model_names = e.args
            return responses.not_found(
                "Models {0} not available for {1}"
                .format(tuple(model_names), context_name))
        except Exception:
            return responses.unknown_error(traceback.format_exc())

    # /v2/scores/enwiki/?models=reverted&revids=456789|4567890
    @bp.route("/v2/scores/<context>/", methods=["GET"])
    @preprocessors.nocache
    @preprocessors.minifiable
    def score_model_revisions_v2(context):
        try:
            score_request = build_score_request(
                scoring_system, request, context)
        except Exception as e:
            return responses.bad_request(str(e))

        return process_score_request(score_request)

    # /v2/scores/enwiki/reverted/?revids=456789|4567890
    @bp.route("/v2/scores/<context>/<model>/", methods=["GET"])
    @preprocessors.nocache
    @preprocessors.minifiable
    def score_revisions_v2(context, model):
        try:
            score_request = build_score_request(
                scoring_system, request, context, model_name=model)
        except Exception as e:
            return responses.bad_request(str(e))

        return process_score_request(score_request)

    # /v2/scores/enwiki/reverted/4567890
    @bp.route("/v2/scores/<context>/<model>/<int:rev_id>/", methods=["GET", "POST"])
    @preprocessors.nocache
    @preprocessors.minifiable
    def score_revision_v2(context, model, rev_id):
        try:
            score_request = build_score_request(
                scoring_system, request, context, rev_id=rev_id,
                model_name=model)
        except Exception as e:
            return responses.bad_request(str(e))

        return process_score_request(score_request)

    return bp



from ...score_request import ScoreRequest
from ..null import Null


def test_null():
    # Make sure we throw no errors.

    collector = Null()
    collector.precache_request(ScoreRequest("foo", [1], {"bar", "derp"}), 100)
    collector.scores_request(ScoreRequest("foo", list(range(50)), {"bar"}), 150)
    collector.datasources_extracted(ScoreRequest("foo", [1], {"bar"}), 10, 25)
    collector.score_processed(ScoreRequest("foo", [1], {"bar"}), 1.1)
    collector.score_timed_out(ScoreRequest("foo", [1], {"bar"}), 15.1)
    collector.score_cache_miss(ScoreRequest("foo", [1], {"derp"}), "derp")
    collector.score_cache_hit(ScoreRequest("foo", [1], {"bar"}), "bar")
    collector.score_errored(ScoreRequest("foo", [1], {"bar"}), "bar")


def test_from_config():
    # Should throw a socket connection error and no others
    config = {
        'metrics_collectors': {
            'null': {
                'class': 'ores.metrics_collectors.Null'
            }
        }
    }
    Null.from_config(config, 'null')



import logging
import time

import revscoring.errors
import stopit

from .. import errors
from ..errors import MissingContext, MissingModels, TimeoutError
from ..metrics_collectors import Null
from ..score_caches import Empty
from ..score_response import ScoreResponse
from ..util import timeout

logger = logging.getLogger(__name__)


class ScoringSystem(dict):

    def __init__(self, context_map, score_cache=None,
                       metrics_collector=None, timeout=None):
        super().__init__()
        self.update(context_map)
        self.score_cache = score_cache or Empty()
        self.metrics_collector = metrics_collector or Null()
        self.timeout = timeout

    def check_context_models(self, request):

        if request.context_name not in self:
            raise MissingContext(request.context_name)

        missing_models = request.model_names - self[request.context_name].keys()
        if len(missing_models) > 0:
            raise MissingModels(request.context_name, missing_models)

    def score(self, request):
        self.check_context_models(request)
        start = time.time()
        logger.debug("Scoring {0}".format(request))

        try:
            response = self._score(request)
        except errors.ScoreProcessorOverloaded:
            self.metrics_collector.score_processor_overloaded(request)
            raise

        duration = time.time() - start
        if not request.precache:
            self.metrics_collector.scores_request(request, duration)
        else:
            self.metrics_collector.precache_request(request, duration)

        return response

    def _score(self, request):
        context = self[request.context_name]
        response = ScoreResponse(context, request)

        # 0. Get model info (if applicable)
        if request.model_info:
            self._build_model_info(request, response)

        # 1. Lookup cached and inprogress scores (Fast IO)
        if not request.include_features:
            # Can't use cached score if we want feature values since those
            # will need to be regenerated
            self._lookup_cached_scores(request, response)
            inprogress_results = self._lookup_inprogress_results(
                request, response)
        else:
            inprogress_results = {}

        # 2. Get missing rev_model sets
        missing_model_set_revs = self._filter_missing_model_set_revs(
            request, response, inprogress_results=inprogress_results)

        # 2.5 Register inprogress work
        self._register_model_set_revs_to_process(
            request, missing_model_set_revs)

        # 3. Extract base datasources for missing models (Slow IO)
        start = time.time()
        root_caches, extraction_errors = self._extract_root_caches(
            request, missing_model_set_revs)
        self.metrics_collector.datasources_extracted(
            request, sum(len(ids) for ids in missing_model_set_revs.values()),
            time.time() - start)

        # 3.5. Record extraction errors
        for rev_id, error in extraction_errors.items():
            for model in request.model_names:
                response.add_error(rev_id, model, error)
                self.metrics_collector.score_errored(request, model)

        # 4. Generate scores (Heavy CPU)
        missing_scores, scoring_errors = self._process_missing_scores(
            request, missing_model_set_revs, root_caches,
            inprogress_results=inprogress_results)
        for rev_id, score_map in missing_scores.items():
            for model_name, model_score in score_map.items():
                response.add_score(rev_id, model_name, model_score['score'])
                if request.include_features:
                    response.add_features(
                        rev_id, model_name, model_score['features'])

            # Store scores in cache
            self._cache_scores(request, rev_id, score_map)

        # 4.5 Record scoring errors
        for rev_id, error in scoring_errors.items():
            for model in request.model_names:
                response.add_error(rev_id, model, error)
                self.metrics_collector.score_errored(request, model)

        return response

    def _build_model_info(self, request, response):
        context = self[request.context_name]
        for model_name in request.model_names:
            response.add_model_info(
                model_name,
                context.format_model_info(model_name, request.model_info))

    def _extract_root_caches(self, request, missing_model_set_revs):
        context = self[request.context_name]

        root_caches = {}
        errors = {}
        for model_set, rev_ids in missing_model_set_revs.items():
            ms_root_caches, ms_errors = context.extract_root_dependency_caches(
                model_set, rev_ids, injection_caches=request.injection_caches)
            root_caches.update(ms_root_caches)
            errors.update(ms_errors)

        return root_caches, errors

    def _process_score_map(self, request, rev_id, model_names, root_cache):
        context = self[request.context_name]

        start = time.time()
        # Runs a timeout function so that we don't get stuck here
        try:
            score_map = timeout(
                context.process_model_scores, model_names, root_cache,
                include_features=request.include_features,
                seconds=self.timeout)
        except revscoring.errors.CaughtDependencyError as e:
            if isinstance(e.exception, stopit.TimeoutException):
                duration = time.time() - start
                self.metrics_collector.score_timed_out(request, duration)
                raise TimeoutError("Timed out after {0} seconds."
                                   .format(duration))
            else:
                raise
        except TimeoutError:
            duration = time.time() - start
            self.metrics_collector.score_timed_out(request, duration)
            raise

        duration = time.time() - start

        logger.debug("Score generated for {0}:{1} in {2} seconds"
                     .format(request.context_name, set(request.model_names),
                             round(duration, 3)))
        self.metrics_collector.score_processed(request, duration)

        return score_map

    def _process_missing_scores(self, request, missing_model_set_revs,
                                root_caches):
        raise NotImplementedError()

    def _filter_missing_model_set_revs(self, request, response,
                                       inprogress_results=None):
        missing_model_set_rev_pairs = self._filter_missing_model_pairs(
            request, response, inprogress_results)
        missing_model_set_revs = {}
        for model_set, rev_id in missing_model_set_rev_pairs:
            if len(model_set) == 0:
                continue
            if model_set in missing_model_set_revs:
                missing_model_set_revs[model_set].append(rev_id)
            else:
                missing_model_set_revs[model_set] = [rev_id]

        return missing_model_set_revs

    def _filter_missing_model_pairs(self, request, response,
                                    inprogress_results):
        for rev_id in request.rev_ids:
            missing_models = request.model_names - (
                set(response.scores.get(rev_id, {}).keys()) |
                set(inprogress_results.get(rev_id, {}).keys()))

            yield frozenset(missing_models), rev_id

    def _register_model_set_revs_to_process(self, request,
                                            missing_model_set_revs):
        return None

    def _cache_scores(self, request, rev_id, score_map):
        for model_name, score_doc in score_map.items():
            self._cache_score(request, rev_id, model_name, score_doc)

    def _cache_score(self, request, rev_id, model_name, score_doc):
        version = self[request.context_name].model_version(model_name)
        injection_cache = request.injection_caches.get(rev_id)
        self.score_cache.store(
            score_doc['score'], request.context_name, model_name,
            rev_id, version=version, injection_cache=injection_cache)

    def _lookup_inprogress_results(self, request, response):
        return {}

    def _lookup_cached_scores(self, request, response):

        for rev_id in request.rev_ids:
            for model_name in request.model_names:
                try:
                    rev_score = self._lookup_cached_score(
                        request, rev_id, model_name)
                    response.add_score(rev_id, model_name, rev_score)
                except KeyError:
                    pass

    def _lookup_cached_score(self, request, rev_id, model_name):
        version = self[request.context_name].model_version(model_name)
        injection_cache = request.injection_caches.get(rev_id)
        try:
            score = self.score_cache.lookup(
                request.context_name, model_name, rev_id,
                version=version, injection_cache=injection_cache)

            logger.debug("Found cached score for {0}"
                         .format(request.format(rev_id, model_name)))

            self.metrics_collector.score_cache_hit(request, model_name)
            return score
        except KeyError:
            self.metrics_collector.score_cache_miss(request, model_name)
            raise

    @classmethod
    def _build_context_map(cls, config, name, section_key="scoring_systems"):
        from ..scoring_context import ScoringContext

        section = config[section_key][name]

        return {name: ScoringContext.from_config(config, name)
                for name in section['scoring_contexts']}

    @classmethod
    def _kwargs_from_config(cls, config, name, section_key="scoring_systems"):
        from ..metrics_collectors import MetricsCollector
        from ..score_caches import ScoreCache

        section = config[section_key][name]

        context_map = cls._build_context_map(
            config, name, section_key=section_key)

        if 'score_cache' in section:
            score_cache = ScoreCache.from_config(config, section['score_cache'])
        else:
            score_cache = None

        if 'metrics_collector' in section:
            metrics_collector = MetricsCollector.from_config(
                config, section['metrics_collector'])
        else:
            metrics_collector = None

        timeout = section.get('timeout')

        return {'context_map': context_map, 'score_cache': score_cache,
                'metrics_collector': metrics_collector, 'timeout': timeout}

    @classmethod
    def from_config(cls, config, name, section_key="scoring_systems"):
        try:
            import yamlconf
        except ImportError:
            raise ImportError("Could not find yamlconf.  This packages is " +
                              "required when using yaml config files.")
        logger.info("Loading ScoreProcessor '{0}' from config.".format(name))
        section = config[section_key][name]
        if 'module' in section:
            return yamlconf.import_module(section['module'])
        elif 'class' in section:
            Class = yamlconf.import_module(section['class'])
            return Class.from_config(config, name)
        else:
            raise RuntimeError("No module or class to load.")



from .metrics_collector import MetricsCollector


class Null(MetricsCollector):
    """
    A non-metrics collector that does nothing.
    """
    def precache_request(self, request, duration):
        pass

    def scores_request(self, request, duration):
        pass

    def datasources_extracted(self, request, rev_id_count, duration):
        pass

    def score_processor_overloaded(self, request):
        pass

    def score_processed(self, request, duration):
        pass

    def score_cache_hit(self, request, model_name):
        pass

    def score_cache_miss(self, request, model_name):
        pass

    def score_errored(self, request, model_name):
        pass

    def score_timed_out(self, request, duration):
        pass

    def precache_scores(self, request, duration):
        pass

    def precache_scoring_error(self, request, status, duration):
        pass

    @classmethod
    def from_config(cls, config, name, section_key="metrics_collectors"):
        """
        metrics_collectors:
        local_logging:
        class: ores.metrics_collectors.Logging
        """
        return cls()



from flask import request
from flask_swaggerui import render_swaggerui

from . import precache
from . import scores
from . import spec


def configure(config, bp, score_processor):

    @bp.route("/v3/", methods=["GET"])
    def v3_index():
        if "spec" in request.args:
            return spec.generate_spec(config)
        else:
            return render_swaggerui(swagger_spec_path="/v3/spec/")

    bp = precache.configure(config, bp, score_processor)
    bp = scores.configure(config, bp, score_processor)
    bp = spec.configure(config, bp, score_processor)

    return bp



from nose.tools import eq_
from revscoring.extractors import OfflineExtractor

from ... import errors
from ...score_caches import LRU
from ...score_request import ScoreRequest
from ...scoring.models import RevIdScorer
from ...scoring_context import ScoringContext
from ..process_pool import ProcessPool
from .util import fakewiki, test_scoring_system, wait_time


def test_score():
    scoring_system = ProcessPool({'fakewiki': fakewiki}, timeout=0.10,
                                 workers=8)

    test_scoring_system(scoring_system)


def test_rev_id_scorer():
    revid = RevIdScorer(version='0.0.1')
    fakewiki = ScoringContext(
        'fakewiki', {'revid': revid}, OfflineExtractor())
    scoring_system = ProcessPool({'fakewiki': fakewiki}, timeout=0.10,
                                 workers=8,
                                 score_cache=LRU(size=10))
    response = scoring_system.score(
        ScoreRequest("fakewiki", [1, 19], ["revid"], model_info=['']))
    print(response.scores, response.errors)
    eq_(response.model_info['revid']['version'], '0.0.1')
    eq_(response.scores[1]['revid']['prediction'], False)
    eq_(response.scores[19]['revid']['prediction'], True)


def test_timeout():
    scoring_system = ProcessPool({'fakewiki': fakewiki}, timeout=0.05)

    response = scoring_system.score(
        ScoreRequest("fakewiki", [1], ['fake'],
                     injection_caches={1: {wait_time: 0.10}}))
    assert 1 in response.errors, str(response.errors)
    assert isinstance(response.errors[1]['fake'], errors.TimeoutError), \
           type(response.errors[1]['fake'])



import os
import re

from setuptools import find_packages, setup

about_path = os.path.join(os.path.dirname(__file__), "ores/about.py")
exec(compile(open(about_path).read(), about_path, "exec"))


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


def requirements(fname):
    """
    Generator to parse requirements.txt file

    Supports bits of extended pip format (git urls)
    """
    with open(fname) as f:
        for line in f:
            match = re.search('#egg=(.*)$', line)
            if match:
                yield match.groups()[0]
            else:
                yield line.strip()


setup(
    name=__name__,  # noqa
    version=__version__,  # noqa
    author=__author__,  # noqa
    author_email=__author_email__,  # noqa
    description=__description__,  # noqa
    url=__url__,  # noqa
    license=__license__,  # noqa
    entry_points={
        'console_scripts': [
            'ores = ores.ores:main',
        ],
    },
    packages=find_packages(),
    include_package_data=True,
    long_description=read('README.md'),
    install_requires=list(requirements("requirements.txt")),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
)



from functools import wraps

from flask import current_app, make_response, request


def minifiable(route):
    @wraps(route)
    def minifiable_route(*args, **kwargs):
        # Change the config
        current_app.config['JSONIFY_PRETTYPRINT_REGULAR'] = \
            False if request.args.get('format') == 'json' else True

        # Generate a response
        response = route(*args, **kwargs)

        # Explicitly return to default now that the result is generated
        current_app.config['JSONIFY_PRETTYPRINT_REGULAR'] = True

        return response

    return minifiable_route


def nocache(route):
    @wraps(route)
    def nocache_route(*args, **kwargs):
        response = make_response(route(*args, **kwargs))
        response.headers['Cache-Control'] = \
            "no-store, no-cache, max-age=0"
        response.headers['Pragma'] = 'no-cache'
        # Unix epoch
        response.headers['Expires'] = 'Thu, 01 Jan 1970 00:00:00 GMT'
        return response

    return nocache_route



from .about import (__author__, __author_email__, __description__, __name__,
                    __url__, __version__)

__all__ = [__name__, __version__, __author__, __author_email__,
           __description__, __url__]



import socket
from contextlib import contextmanager

from nose.tools import eq_, raises

from ...score_request import ScoreRequest
from ..statsd import Statsd


def test_statsd():
    class StatsClient:

        def __init__(self):
            self.messages = []

        def incr(self, name, count=1):
            self.messages.append(("INCR", name, count))

        def timing(self, name, duration):
            self.messages.append(("TIMING", name, duration))

        @contextmanager
        def pipeline(self):
            yield self

    fake_client = StatsClient()

    collector = Statsd(fake_client)
    collector.precache_request(ScoreRequest("foo", [1], {"bar", "derp"}), 100)
    collector.scores_request(ScoreRequest("foo", list(range(50)), {"bar"}), 150)
    collector.datasources_extracted(ScoreRequest("foo", [1], {"bar"}), 10, 25)
    collector.score_processed(ScoreRequest("foo", [1], {"bar"}), 1.1)
    collector.score_timed_out(ScoreRequest("foo", [1], {"bar"}), 15.1)
    collector.score_cache_miss(ScoreRequest("foo", [1], {"derp"}), "derp")
    collector.score_cache_hit(ScoreRequest("foo", [1], {"bar"}), "bar")
    collector.score_errored(ScoreRequest("foo", [1], {"bar"}), "bar")

    eq_(set(fake_client.messages) -
        {('TIMING', 'precache_request.foo.derp', 100000),
         ('TIMING', 'precache_request.foo.bar', 100000),
         ('TIMING', 'precache_request.foo', 100000),
         ('TIMING', 'precache_request', 100000),
         ('TIMING', 'scores_request.foo.bar.50', 150000),
         ('TIMING', 'scores_request.foo.bar', 150000),
         ('TIMING', 'scores_request.foo', 150000),
         ('TIMING', 'scores_request', 150000),
         ('INCR', 'revision_scored.foo.bar', 50),
         ('INCR', 'revision_scored.foo', 50),
         ('INCR', 'revision_scored', 50),
         ('TIMING', 'datasources_extracted.foo.bar.10', 25000),
         ('TIMING', 'datasources_extracted.foo.bar', 25000),
         ('TIMING', 'datasources_extracted.foo', 25000),
         ('TIMING', 'datasources_extracted', 25000),
         ('TIMING', 'score_processed.foo.bar', 1100.0),
         ('TIMING', 'score_processed.foo', 1100.0),
         ('TIMING', 'score_processed', 1100.0),
         ('TIMING', 'score_timed_out.foo.bar', 15100.0),
         ('TIMING', 'score_timed_out.foo', 15100.0),
         ('TIMING', 'score_timed_out', 15100.0),
         ('INCR', 'score_cache_miss.foo.derp', 1),
         ('INCR', 'score_cache_miss.foo', 1),
         ('INCR', 'score_cache_miss', 1),
         ('INCR', 'score_cache_hit.foo.bar', 1),
         ('INCR', 'score_cache_hit.foo', 1),
         ('INCR', 'score_cache_hit', 1),
         ('INCR', 'score_errored.foo.bar', 1),
         ('INCR', 'score_errored.foo', 1),
         ('INCR', 'score_errored', 1)},
        set())


@raises(socket.gaierror)
def test_from_config():
    # Should throw a socket connection error and no others
    config = {
        'metrics_collectors': {
            'wmflabs_statsd': {
                'class': 'ores.metrics_collectors.Statsd',
                'host': 'totally.doesnt.exists.OMG',
                'prefix': 'ores.{hostname}',
                'maxudpsize': 512
            }
        }
    }
    Statsd.from_config(config, 'wmflabs_statsd')



"""
Scores a set of revisions using an ORES api

Usage:
    score_revisions (-h|--help)
    score_revisions <ores-host> <context> <model>...
                    [--input=<path>]
                    [--output=<path>]
                    [--debug]
                    [--verbose]

Options:
    -h --help        Prints this documentation
    <ores-host>      The host name for an ORES instance to use in scoring
    <context>        The name of the wiki to execute model(s) for
    <model>          The name of a model to use in scoring
    --input=<path>   A file containing json blobs with "rev_id" fields
                     [default: <stdin>]
    --output=<path>  A file to write json blobs with scores to
                     [default: <stdout>]
    --debug          Print debug logging
    --verbose        Print out dots to sterr
"""
import json
import logging
import sys

import docopt

from .. import api

logger = logging.getLogger(__name__)


def main(argv=None):
    args = docopt.docopt(__doc__, argv=argv)

    logging.basicConfig(
        level=logging.INFO if not args['--debug'] else logging.DEBUG,
        format='%(asctime)s %(levelname)s:%(name)s -- %(message)s'
    )
    if not args['--debug']:
        logging.getLogger("requests.packages.urllib3").setLevel(logging.WARNING)

    ores_host = args['<ores-host>']
    context = args['<context>']
    model_names = args['<model>']
    if args['--input'] == "<stdin>":
        logger.info("Reading input from <stdin>")
        input = sys.stdin
    else:
        logger.info("Reading input from {0}".format(args['--input']))
        input = open(args['--input'])
    if args['--output'] == "<stdout>":
        logger.info("Writing output to from <stdout>")
        output = sys.stdout
    else:
        logger.info("Writing output to {0}".format(args['--output']))
        output = open(args['--output'], "w")

    verbose = args['--verbose']

    run(ores_host, context, model_names, input, output, verbose)


def run(ores_host, context, model_names, input, output, verbose):
    rev_docs = [json.loads(l) for l in input]
    session = api.Session(ores_host, user_agent="ahalfaker@wikimedia.org")

    rev_ids = [d['rev_id'] for d in rev_docs]
    scores = session.score(context, model_names, rev_ids)

    for rev_doc, score_doc in zip(rev_docs, scores):
        rev_doc['score'] = score_doc
        json.dump(rev_doc, sys.stdout)
        sys.stdout.write("\n")
        if verbose:
            sys.stderr.write(".")
            sys.stderr.flush()


if __name__ == "__main__":
    main()
